# -*- coding: utf-8 -*-
"""
Module for the integration with distutils/setuptools.

Provides functions and classes that enable the management of npm
dependencies for JavaScript sources that lives in Python packages.
"""

from __future__ import absolute_import
import json

from functools import partial
from logging import getLogger

from distutils.command.build import build as BuildCommand
from distutils.errors import DistutilsSetupError

from pkg_resources import Requirement
from pkg_resources import working_set as default_working_set

from calmjs.registry import get
from calmjs.base import BaseModuleRegistry

logger = getLogger(__name__)

# default package definition filename.
JSON_EXTRAS_REGISTRY_KEY = 'calmjs.extras_keys'
CALMJS_MODULE_REGISTRY_FIELD = 'calmjs_module_registry'
CALMJS_MODULE_REGISTRY_TXT = 'calmjs_module_registry.txt'
DEFAULT_JSON = 'default.json'
EXTRAS_CALMJS_FIELD = 'extras_calmjs'
EXTRAS_CALMJS_JSON = 'extras_calmjs.json'
DEP_KEYS = ('dependencies', 'devDependencies')
TEST_REGISTRY_NAME_SUFFIX = '.tests'


def is_json_compat(value):
    """
    Check that the value is either a JSON decodable string or a dict
    that can be encoded into a JSON.

    Raises ValueError when validation fails.
    """

    try:
        value = json.loads(value)
    except ValueError as e:
        raise ValueError('JSON decoding error: ' + str(e))
    except TypeError:
        # Check that the value can be serialized back into json.
        try:
            json.dumps(value)
        except TypeError as e:
            raise ValueError(
                'must be a JSON serializable object: ' + str(e))

    if not isinstance(value, dict):
        raise ValueError(
            'must be specified as a JSON serializable dict or a '
            'JSON deserializable string'
        )

    return True


def validate_json_field(dist, attr, value):
    """
    Check for json validity.
    """

    try:
        is_json_compat(value)
    except ValueError as e:
        raise DistutilsSetupError("%r %s" % (attr, e))

    return True


def validate_line_list(dist, attr, value):
    """
    Validate that the value is compatible
    """

    # does not work as reliably in Python 2.
    if isinstance(value, str):
        value = value.split()
    value = list(value)

    try:
        check = (' '.join(value)).split()
        if check == value:
            return True
    except Exception:
        pass
    raise DistutilsSetupError("%r must be a list of valid identifiers" % attr)


def write_json_file(argname, cmd, basename, filename):
    """
    Write JSON captured from the defined argname into the package's
    egg-info directory using the specified filename.
    """

    value = getattr(cmd.distribution, argname, None)

    if isinstance(value, dict):
        value = json.dumps(
            value, indent=4, sort_keys=True, separators=(',', ': '))

    cmd.write_or_delete_file(argname, filename, value, force=True)


def write_line_list(argname, cmd, basename, filename):
    """
    Write out the retrieved value as list of lines.
    """

    values = getattr(cmd.distribution, argname, None)
    if isinstance(values, list):
        values = '\n'.join(values)
    cmd.write_or_delete_file(argname, filename, values, force=True)


def find_pkg_dist(pkg_name, working_set=None):
    """
    Locate a package's distribution by its name.
    """

    working_set = working_set or default_working_set
    req = Requirement.parse(pkg_name)
    return working_set.find(req)


def convert_package_names(package_names):
    """
    Convert package names, which can be a string of a number of package
    names or requirements separated by spaces.
    """

    results = []
    errors = []

    for name in (
            package_names.split()
            if hasattr(package_names, 'split') else package_names):
        try:
            Requirement.parse(name)
        except ValueError:
            errors.append(name)
        else:
            results.append(name)

    return results, errors


def pkg_names_to_dists(pkg_names, working_set=None):
    working_set = working_set or default_working_set
    return [dist for dist in (find_pkg_dist(
        pkg_name, working_set=working_set) for pkg_name in pkg_names) if dist]


def find_packages_requirements_dists(pkg_names, working_set=None):
    """
    Return the entire list of dependency requirements, reversed from the
    bottom.
    """

    working_set = working_set or default_working_set
    requirements = [
        r for r in (Requirement.parse(req) for req in pkg_names)
        if working_set.find(r)
    ]
    return list(reversed(working_set.resolve(requirements)))


def find_packages_parents_requirements_dists(pkg_names, working_set=None):
    """
    Leverages the `find_packages_requirements_dists` but strip out the
    distributions that matches pkg_names.
    """

    dists = []
    # opting for a naive implementation
    targets = set(pkg_names)
    for dist in find_packages_requirements_dists(pkg_names, working_set):
        if dist.project_name in targets:
            continue
        dists.append(dist)
    return dists


def read_dist_egginfo_json(dist, filename=DEFAULT_JSON):
    """
    Safely get a json within an egginfo from a distribution.
    """

    # use the given package's distribution to acquire the json file.
    if not dist.has_metadata(filename):
        logger.debug("no '%s' for '%s'", filename, dist)
        return

    try:
        result = dist.get_metadata(filename)
    except IOError:
        logger.error("I/O error on reading of '%s' for '%s'.", filename, dist)
        return

    try:
        obj = json.loads(result)
    except (TypeError, ValueError):
        logger.error(
            "the '%s' found in '%s' is not a valid json.", filename, dist)
        return

    logger.debug("found '%s' for '%s'.", filename, dist)
    return obj


def read_egginfo_json(pkg_name, filename=DEFAULT_JSON, working_set=None):
    """
    Read json from egginfo of a package identified by `pkg_name` that's
    already installed within the current Python environment.
    """

    working_set = working_set or default_working_set
    dist = find_pkg_dist(pkg_name, working_set=working_set)
    return read_dist_egginfo_json(dist, filename)


def read_dist_line_list(dist, filename):
    if not dist.has_metadata(filename):
        return []

    try:
        result = dist.get_metadata(filename)
    except IOError:
        # not as critical as egginfo json, as typical usage these can
        # be easily overridden
        logger.warning("I/O error on reading of '%s' for '%s'", filename, dist)
        return []

    return result.split()


def flatten_dist_egginfo_json(
        source_dists, filename=DEFAULT_JSON, dep_keys=DEP_KEYS,
        working_set=None):
    """
    Flatten a distribution's egginfo json, with the depended keys to be
    flattened.

    Originally this was done for this:

    Resolve a distribution's (dev)dependencies through the working set
    and generate a flattened version package.json, returned as a dict,
    from the resolved distributions.

    Default working set is the one from pkg_resources.

    The generated package.json dict is done by grabbing all package.json
    metadata from all parent Python packages, starting from the highest
    level and down to the lowest.  The current distribution's
    dependencies will be layered on top along with its other package
    information.  This has the effect of child packages overriding
    node/npm dependencies which is by the design of this function.  If
    nested dependencies are desired, just rely on npm only for all
    dependency management.

    Flat is better than nested.
    """

    working_set = working_set or default_working_set
    obj = {}

    # TODO figure out the best way to explicitly report back to caller
    # how the keys came to be (from which dist).  Perhaps create a
    # detailed function based on this, retain this one to return the
    # distilled results.

    depends = {dep: {} for dep in dep_keys}

    # Go from the earliest package down to the latest one, as we will
    # flatten children's d(evD)ependencies on top of parent's.
    for dist in source_dists:
        obj = read_dist_egginfo_json(dist, filename)
        if not obj:
            continue

        logger.debug("merging '%s' for required '%s'", filename, dist)
        for dep in dep_keys:
            depends[dep].update(obj.get(dep, {}))

    if obj is None:
        # top level object does not have egg-info defined
        return depends

    for dep in dep_keys:
        # filtering out all the nulls.
        obj[dep] = {k: v for k, v in depends[dep].items() if v is not None}

    return obj


def flatten_egginfo_json(
        pkg_names, filename=DEFAULT_JSON, dep_keys=DEP_KEYS, working_set=None):
    """
    A shorthand calling convention where the package name is supplied
    instead of a distribution.

    Originally written for this:

    Generate a flattened package.json with packages `pkg_names` that are
    already installed within the current Python environment (defaults
    to the current global working_set which should have been set up
    correctly by pkg_resources).
    """

    working_set = working_set or default_working_set
    # Ensure only grabbing packages that exists in working_set
    dists = find_packages_requirements_dists(
        pkg_names, working_set=working_set)
    return flatten_dist_egginfo_json(
        dists, filename=filename, dep_keys=dep_keys, working_set=working_set)


def build_helpers_egginfo_json(
        json_field, json_key_registry, json_filename=None):
    """
    Return a tuple of functions that will provide the usage of the
    JSON egginfo based around the provided field.
    """

    json_filename = (
        json_field + '.json' if json_filename is None else json_filename)

    # Default calmjs core implementation specific functions, to be used by
    # integrators intended to use this as a distribution.

    def get_extras_json(pkg_names, working_set=None):
        """
        Only extract the extras_json information for the given packages
        'pkg_names'.
        """

        working_set = working_set or default_working_set
        dep_keys = set(get(json_key_registry).iter_records())
        dists = pkg_names_to_dists(pkg_names, working_set=working_set)
        return flatten_dist_egginfo_json(
            dists, filename=json_filename,
            dep_keys=dep_keys, working_set=working_set
        )

    def _flatten_extras_json(pkg_names, find_dists, working_set):
        # registry key must be explicit here as it was designed for this.
        dep_keys = set(get(json_key_registry).iter_records())
        dists = find_dists(pkg_names, working_set=working_set)
        return flatten_dist_egginfo_json(
            dists, filename=json_filename,
            dep_keys=dep_keys, working_set=working_set
        )

    def flatten_extras_json(pkg_names, working_set=None):
        """
        Traverses through the dependency graph of packages 'pkg_names'
        and flattens all the egg_info json information
        """

        working_set = working_set or default_working_set
        return _flatten_extras_json(
            pkg_names, find_packages_requirements_dists, working_set)

    def flatten_parents_extras_json(pkg_names, working_set=None):
        """
        Traverses through the dependency graph of packages 'pkg_names'
        and flattens all the egg_info json information for parents of
        the specified packages.
        """

        working_set = working_set or default_working_set
        return _flatten_extras_json(
            pkg_names, find_packages_parents_requirements_dists, working_set)

    write_extras_json = partial(write_json_file, json_field)

    return (
        get_extras_json,
        flatten_extras_json,
        flatten_parents_extras_json,
        write_extras_json,
    )


(get_extras_calmjs, flatten_extras_calmjs, flatten_parents_extras_calmjs,
    write_extras_calmjs) = build_helpers_egginfo_json(
        EXTRAS_CALMJS_FIELD, JSON_EXTRAS_REGISTRY_KEY)


def build_helpers_module_registry_dependencies(registry_name='calmjs.module'):
    """
    Return a tuple of funtions that will provide the functions that
    return the relevant sets of module registry records based on the
    dependencies defined for the provided packages.
    """

    def get_module_registry_dependencies(
            pkg_names, registry_name=registry_name, working_set=None):
        """
        Get dependencies for the given package names from module
        registry identified by registry name.

        For the given packages 'pkg_names' and the registry identified
        by 'registry_name', resolve the exported location for just the
        package.
        """

        working_set = working_set or default_working_set
        registry = get(registry_name)
        if not isinstance(registry, BaseModuleRegistry):
            return {}
        result = {}
        for pkg_name in pkg_names:
            result.update(registry.get_records_for_package(pkg_name))
        return result

    def _flatten_module_registry_dependencies(
            pkg_names, registry_name, find_dists, working_set):
        """
        Flatten dependencies for the given package names from module
        registry identified by registry name using the find_dists
        function on the given working_set.

        For the given packages 'pkg_names' and the registry identified
        by 'registry_name', resolve and flatten all the exported
        locations.
        """

        result = {}
        registry = get(registry_name)
        if not isinstance(registry, BaseModuleRegistry):
            return result

        dists = find_dists(pkg_names, working_set=working_set)
        for dist in dists:
            result.update(registry.get_records_for_package(dist.project_name))

        return result

    def flatten_module_registry_dependencies(
            pkg_names, registry_name=registry_name, working_set=None):
        """
        Flatten dependencies for the specified packages from the module
        registry identified by registry name.

        For the given packages 'pkg_names' and the registry identified
        by 'registry_name', resolve and flatten all the exported
        locations.
        """

        working_set = working_set or default_working_set
        return _flatten_module_registry_dependencies(
            pkg_names, registry_name, find_packages_requirements_dists,
            working_set)

    def flatten_parents_module_registry_dependencies(
            pkg_names, registry_name=registry_name, working_set=None):
        """
        Flatten dependencies for the parents of the specified packages
        from the module registry identified by registry name.

        For the given packages 'pkg_names' and the registry identified
        by 'registry_name', resolve and flatten all the exported
        locations.
        """

        working_set = working_set or default_working_set
        return _flatten_module_registry_dependencies(
            pkg_names, registry_name, find_packages_parents_requirements_dists,
            working_set)

    return (
        get_module_registry_dependencies,
        flatten_module_registry_dependencies,
        flatten_parents_module_registry_dependencies,
    )


(get_module_registry_dependencies, flatten_module_registry_dependencies,
    flatten_parents_module_registry_dependencies) = (
        build_helpers_module_registry_dependencies())


def _uniq(items):
    check = set()
    return [i for i in items if not (i in check or check.add(i))]


def build_helpers_module_registry_name(
        registry_field, registry_field_txt=None):

    registry_field_txt = (
        registry_field + '.txt' if registry_field_txt is None else
        registry_field_txt
    )

    def get_module_registry_names(pkg_names, working_set=None):
        """
        Get names of module registries registered for package names.

        For the given packages 'pkg_names', retrieve the list of module
        registries explicitly declared for usage.
        """

        ws = working_set or default_working_set
        result = []
        for dist in pkg_names_to_dists(pkg_names, working_set=ws):
            result.extend(read_dist_line_list(dist, registry_field_txt))
        return _uniq(result)

    def flatten_module_registry_names(pkg_names, working_set=None):
        """
        Flatten all names of module registries registered for package names.

        For the given packages 'pkg_names' and its dependencies, retrieve
        the list of module registries explicitly declared for usage.
        """

        ws = working_set or default_working_set
        result = []
        for dist in find_packages_requirements_dists(
                pkg_names, working_set=ws):
            result.extend(read_dist_line_list(dist, registry_field_txt))
        return _uniq(result)

    write_module_registry_names = partial(write_line_list, registry_field)

    return (
        get_module_registry_names,
        flatten_module_registry_names,
        write_module_registry_names,
    )


(get_module_registry_names, flatten_module_registry_names,
    write_module_registry_names) = build_helpers_module_registry_name(
        CALMJS_MODULE_REGISTRY_FIELD)


# These depend on the artifact registry that is defined in the artifact
# module.

def has_calmjs_artifact_declarations(cmd, registry_name='calmjs.artifacts'):
    """
    For a distutils command to verify that the artifact build step is
    possible.
    """

    return any(get(registry_name).iter_records_for(
        cmd.distribution.get_name()))


def build_calmjs_artifacts(dist, key, value, cmdclass=BuildCommand):
    """
    Trigger the artifact build process through the setuptools.
    """

    if value is not True:
        return

    build_cmd = dist.get_command_obj('build')
    if not isinstance(build_cmd, cmdclass):
        logger.error(
            "'build' command in Distribution is not an instance of "
            "'%s:%s' (got %r instead)",
            cmdclass.__module__, cmdclass.__name__, build_cmd)
        return

    build_cmd.sub_commands.append((key, has_calmjs_artifact_declarations))



# -*- coding: utf-8 -*-
"""
calmjs.base

This module contains various base classes from which other parts of this
framework will inherit/extend from.
"""

from __future__ import absolute_import
import os

import errno
import json
from os import getcwd
from os.path import dirname
from os.path import isdir
from os.path import join
from os.path import pathsep
from os.path import realpath
from os.path import sep

from collections import OrderedDict
try:
    from collections.abc import MutableMapping
except ImportError:  # pragma: no cover
    from collections import MutableMapping
from logging import getLogger
from pkg_resources import Distribution
from pkg_resources import working_set
from pkg_resources import safe_name

from calmjs.utils import which
from calmjs.utils import finalize_env
from calmjs.utils import fork_exec
from calmjs.utils import raise_os_error

NODE_PATH = 'NODE_PATH'
NODE_MODULES = 'node_modules'
# the usual path to binary within node modules.
NODE_MODULES_BIN = '.bin'
NODE = 'node'

logger = getLogger(__name__)
_marker = object()


def _import_module(module_name):
    return __import__(module_name, fromlist=['__name__'], level=0)


def _check_isdir_assign_key(d, key, value, error_msg=None):
    if isdir(value):
        d[key] = value
    else:
        extra_info = '' if not error_msg else '; ' + error_msg
        logger.error(
            "not manually setting '%s' to '%s' as it not a directory%s",
            key, value, extra_info
        )


def _get_exec_binary(binary, kw):
    """
    On win32, the subprocess module can only reliably resolve the
    target binary if it's actually a binary; as for a Node.js script
    it seems to only work iff shell=True was specified, presenting
    a security risk.  Resolve the target manually through which will
    account for that.

    The kw argument is the keyword arguments that will be passed into
    whatever respective subprocess.Popen family of methods.  The PATH
    environment variable will be used if available.
    """

    binary = which(binary, path=kw.get('env', {}).get('PATH'))
    if binary is None:
        raise_os_error(errno.ENOENT)
    return binary


class PackageKeyMapping(MutableMapping):
    """
    A mapping where keys are pkg_resources.Distribution.project_names
    mapping to some value.  As project_names are stored, lookup using
    the de-normalized values must go through the safe_name normalization
    such that the resolution will work as expected.
    """

    def __init__(self, *a, **kw):
        self.__map = {}
        # calling self.update instead to use the defined methods that
        # will call normalize.
        self.update(*a, **kw)

    def normalize(self, key):
        return safe_name(key)

    def __getitem__(self, key):
        return self.__map[self.normalize(key)]

    def __setitem__(self, key, value):
        if isinstance(key, Distribution):
            self.__map[key.project_name] = value
        else:
            self.__map[self.normalize(key)] = value

    def __delitem__(self, key):
        self.__map.__delitem__(self.normalize(key))

    def __iter__(self):
        return iter(self.__map)

    def __len__(self):
        return len(self.__map)

    def __contains__(self, key):
        return self.normalize(key) in self.__map

    def __repr__(self):
        return repr(self.__map)


class BaseRegistry(object):
    """
    A base registry implementation that make use of ``pkg_resources``
    entry points for its definitions.
    """

    def __init__(self, registry_name, *a, **kw):
        """
        Arguments:

        registry_name
            The name of this registry.
        """

        # The container for the resolved item.
        self.records = OrderedDict()
        self.registry_name = registry_name
        _working_set = kw.pop('_working_set', working_set)
        self.raw_entry_points = [] if _working_set is None else list(
            _working_set.iter_entry_points(self.registry_name))
        self._init(*a, **kw)

    def _init(self, *a, **kw):
        """
        Subclasses can completely override this for its instantiation,
        or it can override this by having its subclass override
        _init_entry_point and then have this method call

            self._init_entry_points(self.raw_entry_points)

        to initialize all entry points during instantiation.
        """

    def _init_entry_points(self, entry_points):
        """
        Default initialization loop.
        """

        logger.debug(
            "registering %d entry points for registry '%s'",
            len(entry_points), self.registry_name,
        )
        for entry_point in entry_points:
            try:
                logger.debug(
                    "registering entry point '%s' from '%s'",
                    entry_point, entry_point.dist,
                )
                self._init_entry_point(entry_point)
            except ImportError:
                logger.warning(
                    'ImportError: %s not found; skipping registration',
                    entry_point.module_name)
            except Exception:
                logger.exception(
                    "registration of entry point '%s' from '%s' to registry "
                    "'%s' failed with the following exception",
                    entry_point, entry_point.dist, self.registry_name,
                )

    def _init_entry_point(self, entry_point):
        """
        Default does nothing.
        """

    def get_record(self, name):
        raise NotImplementedError

    def get(self, name):
        return self.get_record(name)

    def iter_records(self):
        raise NotImplementedError


class BasePkgRefRegistry(BaseRegistry):
    """
    A common base registry that deals with references for data within
    packages.
    """

    def __init__(self, registry_name, *a, **kw):
        super(BasePkgRefRegistry, self).__init__(registry_name, *a, **kw)
        self.package_module_map = PackageKeyMapping()
        self.register_entry_points(self.raw_entry_points)

    def register_entry_points(self, entry_points):
        """
        Register all entry_points provided by the list, if and only if
        the associated module can be imported.

        Arguments:

        entry_points
            a list of entry_points to be registered and activated into
            this registry instance.
        """

        return self._init_entry_points(entry_points)

    def _init_entry_point(self, entry_point):
        return self.register_entry_point(entry_point)

    def register_entry_point(self, entry_point):
        raise NotImplementedError

    def _dist_to_package_module_map(self, entry_point):
        # TODO fix the two following logging statement, likely indicates
        # that a proper cleanup/refactoring is needed for records in
        # this registry type.
        if entry_point.dist is None:
            # it's probably manually added not through the standard
            # setuptools procedures.
            logger.warning(
                "manually registering entry_point '%s' without associated "
                "distribution to registry '%s'",
                entry_point, self.registry_name,
            )
            return []  # just a dummy unconnected value.
        else:
            logger.debug(
                "registering entry_point '%s' from '%s' to registry '%s'",
                entry_point, entry_point.dist, self.registry_name,
            )
            if entry_point.dist.project_name not in self.package_module_map:
                self.package_module_map[entry_point.dist.project_name] = []
            return self.package_module_map[entry_point.dist.project_name]

    def store_records_for_package(self, entry_point, records):
        """
        Store the records in a way that permit lookup by package
        """

        # If provided records already exist in the module mapping list,
        # it likely means that a package declared multiple keys for the
        # same package namespace; while normally this does not happen,
        # this default implementation make no assumptions as to whether
        # or not this is permitted.
        pkg_module_records = self._dist_to_package_module_map(entry_point)
        pkg_module_records.extend(records)

    def iter_records(self):
        """
        Iterates through the records.
        """

        for item in self.records.items():
            yield item


class BaseModuleRegistry(BasePkgRefRegistry):
    """
    Extending off the BasePkgRefRegistry, ensure that there is a
    registration step that takes place that will verify the existence
    of the target.
    """

    def register_entry_point(self, entry_point):
        """
        Register a lone entry_point

        Will raise ImportError if the entry_point leads to an invalid
        import.
        """

        module = _import_module(entry_point.module_name)
        self._register_entry_point_module(entry_point, module)

    def _register_entry_point_module(self, entry_point, module):
        """
        Private method that registers an entry_point with a provided
        module.
        """

        records_map = self._map_entry_point_module(entry_point, module)
        self.store_records_for_package(entry_point, list(records_map.keys()))

        for module_name, records in records_map.items():
            if module_name in self.records:
                logger.info(
                    "module '%s' was already declared in registry '%s'; "
                    "applying new records on top.",
                    module_name, self.registry_name,
                )
                logger.debug("overwriting keys: %s", sorted(
                    set(self.records[module_name].keys()) &
                    set(records.keys())
                ))
                self.records[module_name].update(records)
            else:
                logger.debug(
                    "adding records for module '%s' to registry '%s'",
                    module_name, self.registry_name,
                )
                self.records[module_name] = records

    def _map_entry_point_module(self, entry_point, module):
        """
        Subclass need to implement this.

        The implementation is to return a map (dict) of the name of the
        module(s) with their mapped source files.
        """

        raise NotImplementedError

    def get_record(self, name):
        """
        Get a record by name
        """

        result = {}
        result.update(self.records.get(name, {}))
        return result

    def get_records_for_package(self, package_name):
        """
        Get all records identified by package.
        """

        names = self.package_module_map.get(package_name, [])
        result = {}
        for name in names:
            result.update(self.get_record(name))
        return result


class BaseChildModuleRegistry(BaseModuleRegistry):
    """
    To enable more generic support of child module registries, where the
    records depend on a parent registry, this class provides the basic
    framework to do so.
    """

    def __init__(self, registry_name, *a, **kw):
        # TODO whenever there is time to move the BaseRegistry to a
        # bootstrap module of some sort (that will break tests that
        # override calmjs.base.working_set if done naively), and have
        # the calmjs.registry.Registry inherit from that (and this
        # module also to maintain the BaseRegistry import location,
        # this import should be moved to the top
        from calmjs.registry import get
        # resolve parent before the parent class, as the construction
        # should be able to reference this parent.
        parent_name = self.resolve_parent_registry_name(registry_name)
        _parent = kw.pop('_parent', NotImplemented)
        if _parent is NotImplemented:
            self.parent = get(parent_name)
        else:
            self.parent = _parent

        if not self.parent:
            raise ValueError(
                "could not construct child module registry '%s' as its "
                "parent registry '%s' could not be found" % (
                    registry_name, parent_name)
            )
        super(BaseChildModuleRegistry, self).__init__(registry_name, *a, **kw)

    def resolve_parent_registry_name(self, registry_name, suffix):
        """
        Subclasses should override to specify the default suffix, as the
        invocation is done without a suffix.
        """

        if not registry_name.endswith(suffix):
            raise ValueError(
                "child module registry name defined with invalid suffix "
                "('%s' does not end with '%s')" % (registry_name, suffix))
        return registry_name[:-len(suffix)]


class BaseExternalModuleRegistry(BasePkgRefRegistry):
    """
    A registry for storing references to scripts sourced from JavaScript
    or Node.js module manager, i.e. node_modules.

    The intent of this registry is to have register paths that point to
    targets managed in those external systems via the key portion of the
    entry points.  The interpretation of the values will be specific to
    the implementation.

    At the very least, a value of true can simply be used to denote that
    a given package will want to do something with it.

    Another use case might be for the referencing of prebuilt artifacts
    and optionally give them an alias (however, it must be one that is
    compatible to the Python module name scheme given the limitations
    due to the entry point format).
    """

    # TODO decide on providing support for a handler, if an attribute to
    # some provided module was provided instead, so that it will resolve
    # the actual name with that callable - this means supplimentary info
    # can be provided, which could be provided via something like
    # calmjs_extras.

    def register_entry_point(self, entry_point):
        # for storing a mapping from the Python module that the
        # declarations to the JavaScript modules themselves.
        paths = self.process_entry_point(entry_point)
        self.store_record(entry_point, paths)
        self.store_records_for_package(entry_point, paths)

    def store_record(self, entry_point, paths):
        # the record is stored as an inverse mapping to the set of
        # paths that called on that module_name.
        self.records[entry_point.module_name] = self.records.get(
            entry_point.module_name, set()).union(paths)

    def process_entry_point(self, entry_point):
        """
        The default implementation simply return the entry point to a
        path.

        Result type should be an iterable.
        """

        return [entry_point.name]

    def get_record(self, name):
        """
        Get a record for the registered name, which will be a set of
        matching desired "module names" for the given path.
        """

        return set().union(self.records.get(name, set()))

    def get_records_for_package(self, package_name):
        """
        Get all records identified by package.
        """

        result = []
        result.extend(self.package_module_map.get(package_name))
        return result


class BaseDriver(object):
    """
    The Node.js interfacing base driver class.

    Classes under the calmjs framework that make use of Node.js or
    Node.js binaries should make this as the base class to make
    accessing the Node.js environment in a manner consistent with the
    framework, where various defined attributes and helper methods that
    makes use of those will make interfacing with the Node.js
    environment be under a consistently managed scheme.

    Helper methods such as _exec will ensure the binary is executed
    with the right arguments (via _gen_call_kws).  The dump/dumps method
    invokes the underlying functions of the same name from the json
    module with the attributes defined to ensure human readability by
    default.  Finally, join_cwd joins a target path with the working
    directory defined for instances of this, so that target path can be
    accessed in a more explicit way.
    """

    def __init__(self, node_path=None, env_path=None, working_dir=None,
                 indent=4, separators=(',', ': '), description=None):
        """
        Optional Arguments (defaults to None when not stated):

        node_path
            Overrides NODE_PATH environment variable for calling out.
        env_path
            Extra directory that will be assigned to the environment's
            PATH variable, so that that will be used first to look up a
            binary for exec calls.
        working_dir
            The working directory where the driver will operate at.
        indent
            JSON indentation level.  Defaults to 4.
        separators
            Set as a workaround to remove trailing spaces.  Should be
            left as is (',', ': ').
        description
            A description for this driver.  Optional string.
        """

        self.node_path = node_path or os.environ.get(NODE_PATH)
        self.env_path = env_path
        self.working_dir = working_dir
        self.indent = indent
        self.separators = separators
        self.binary = None
        self.description = description

    def which(self):
        """
        Figure out which binary this will execute.

        Returns None if the binary is not found.
        """

        if self.binary is None:
            return None

        return which(self.binary, path=self.env_path)

    def find_node_modules_basedir(self):
        """
        Find all node_modules directories configured to be accessible
        through this driver instance.

        This is typically used for adding the direct instance, and does
        not traverse the parent directories like what Node.js does.

        Returns a list of directories that contain a 'node_modules'
        directory.
        """

        paths = []

        # First do the working dir.
        local_node_path = self.join_cwd(NODE_MODULES)
        if isdir(local_node_path):
            paths.append(local_node_path)

        # do the NODE_PATH environment variable last, as Node.js seem to
        # have these resolving just before the global.
        if self.node_path:
            paths.extend(self.node_path.split(pathsep))

        return paths

    def which_with_node_modules(self):
        """
        Which with node_path and node_modules
        """

        if self.binary is None:
            return None

        # first, log down the pedantic things...
        if isdir(self.join_cwd(NODE_MODULES)):
            logger.debug(
                "'%s' instance will attempt to locate '%s' binary from "
                "%s%s%s%s%s, located through the working directory",
                self.__class__.__name__, self.binary,
                self.join_cwd(), sep, NODE_MODULES, sep, NODE_MODULES_BIN,
            )
        if self.node_path:
            logger.debug(
                "'%s' instance will attempt to locate '%s' binary from "
                "its %s of %s",
                self.__class__.__name__, self.binary,
                NODE_PATH, self.node_path,
            )

        paths = self.find_node_modules_basedir()
        whichpaths = pathsep.join(join(p, NODE_MODULES_BIN) for p in paths)

        if paths:
            logger.debug(
                "'%s' instance located %d possible paths to the '%s' binary, "
                "which are %s",
                self.__class__.__name__, len(paths), self.binary, whichpaths,
            )

        return which(self.binary, path=whichpaths)

    @classmethod
    def create(cls):
        """
        Freeze an instance to the current working directory and its
        related environmental settings.
        """

        inst = cls()
        inst._set_env_path_with_node_modules()
        return inst

    def _set_env_path_with_node_modules(self):
        """
        Attempt to locate and set the paths to the binary with the
        working directory defined for this instance.
        """

        modcls_name = ':'.join((
            self.__class__.__module__, self.__class__.__name__))

        if self.binary is None:
            raise ValueError(
                "binary undefined for '%s' instance" % modcls_name)

        logger.debug(
            "locating '%s' node binary for %s instance...",
            self.binary, modcls_name,
        )

        default = self.which()
        if default is not None:
            logger.debug(
                "found '%s'; "
                "not modifying PATH environment variable in instance of '%s'.",
                realpath(default), modcls_name)
            return True

        target = self.which_with_node_modules()

        if target:
            # Only setting the path specific for the binary; side effect
            # will be whoever else borrowing the _exec in here might not
            # get the binary they want.  That's why it's private.
            self.env_path = dirname(target)
            logger.debug(
                "located '%s' binary at '%s'; setting PATH environment "
                "variable for '%s' instance.",
                self.binary, self.env_path, modcls_name
            )
            return True
        else:
            logger.debug(
                "Unable to locate '%s'; not modifying PATH environment "
                "variable for instance of '%s'.",
                self.binary, modcls_name
            )
            return False

    def _gen_call_kws(self, **env):
        kw = {}
        if self.node_path is not None:
            env[NODE_PATH] = self.node_path
        if self.env_path is not None:
            # Initial assignment with check
            _check_isdir_assign_key(env, 'PATH', self.env_path)
            # then append the rest of it
            env['PATH'] = pathsep.join([
                env.get('PATH', ''), os.environ.get('PATH', '')])
        if self.working_dir:
            _check_isdir_assign_key(
                kw, 'cwd', self.working_dir,
                error_msg="current working directory left as default")
        kw['env'] = finalize_env(env)
        return kw

    def _get_exec_binary(self, kw):
        """
        This wraps the base function for BaseDriver classes; should only
        be called by this instance's binary execution methods as they
        will be also invoke _gen_call_kws to pass into the call_kw
        argument.

        This is different from the `which` method as the underlying
        function will raise the correct exception as if the underlying
        Popen method was called directly with the self.binary argument,
        while `which` will simply return None.
        """

        return _get_exec_binary(self.binary, kw)

    def _exec(self, binary, stdin='', args=(), env={}):
        """
        Executes the binary using stdin and args with environment
        variables.

        Returns a tuple of stdout, stderr.  Format determined by the
        input text (either str or bytes), and the encoding of str will
        be determined by the locale this module was imported in.
        """

        call_kw = self._gen_call_kws(**env)
        call_args = [self._get_exec_binary(call_kw)]
        call_args.extend(args)
        return fork_exec(call_args, stdin, **call_kw)

    @property
    def cwd(self):
        return self.working_dir or getcwd()

    def dump(self, blob, stream):
        """
        Call json.dump with the attributes of this instance as
        arguments.
        """

        json.dump(
            blob, stream, indent=self.indent, sort_keys=True,
            separators=self.separators,
        )

    def dumps(self, blob):
        """
        Call json.dumps with the attributes of this instance as
        arguments.
        """

        return json.dumps(
            blob, indent=self.indent, sort_keys=True,
            separators=self.separators,
        )

    def join_cwd(self, path=None):
        """
        Join the path with the current working directory.  If it is
        specified for this instance of the object it will be used,
        otherwise rely on the global value.
        """

        if self.working_dir:
            logger.debug(
                "'%s' instance 'working_dir' set to '%s' for join_cwd",
                type(self).__name__, self.working_dir,
            )
            cwd = self.working_dir
        else:
            cwd = getcwd()
            logger.debug(
                "'%s' instance 'working_dir' unset; "
                "default to process '%s' for join_cwd",
                type(self).__name__, cwd,
            )

        if path:
            return join(cwd, path)
        return cwd


class BaseLoaderPluginRegistry(BaseRegistry):

    def _init_entry_point(self, entry_point):
        try:
            cls = entry_point.load()
        except ImportError:
            logger.warning(
                "registry '%s' failed to load loader plugin handler for "
                "entry point '%s'", self.registry_name, entry_point,
            )
            return

        if not issubclass(cls, BaseLoaderPluginHandler):
            logger.warning(
                "entry point '%s' does not lead to a valid loader plugin "
                "handler class", entry_point
            )
            return

        inst = cls(self, entry_point.name)

        if entry_point.name in self.records:
            old = type(self.records[entry_point.name])
            logger.warning(
                "loader plugin handler for '%s' was already registered to "
                "an instance of '%s:%s'; '%s' will now override this "
                "registration",
                entry_point.name, old.__module__, old.__name__, entry_point
            )
        self.records[entry_point.name] = inst

    def to_plugin_name(self, value):
        """
        Find the plugin name from the provided value
        """

        return value.split('!', 1)[0].split('?', 1)[0]

    def get_record(self, name):
        # it is possible for subclasses to provide a fallback lookup on
        # "common" registries through the registry framework, e.g.
        # calmjs.registry.get('some.plugin.reg').get_record(name)
        return self.records.get(self.to_plugin_name(name))


class BaseLoaderPluginHandler(object):
    """
    The base loaderplugin handler class provides only the stub methods;
    for a more concrete implementation, refer to the loaderplugin module
    for the subclass.
    """

    def __init__(self, registry, name=None):
        """
        The LoaderPluginRegistry will try to construct the instance and
        pass itself into the constructor; leaving this as the default
        will enable specific plugins to load further plugins should the
        input modname has more loader plugin strings.
        """

        self.registry = registry
        self.name = name

    def modname_source_to_target(
            self, toolchain, spec, modname, source):
        """
        This is called by the Toolchain for modnames that contain a '!'
        as that signifies a loaderplugin syntax.  This will be used by
        the toolchain (which will also be supplied as the first argument)
        to resolve the copy target, which must be a path relative to the
        spec[WORKING_DIR].

        If the provided modname points contains a chain of loaders, the
        registry associated with this handler instance will be used to
        resolve the subsequent handlers until none are found, which that
        handler will be used to return this result.
        """

        stripped_modname = self.unwrap(modname)
        chained = (
            self.registry.get_record(stripped_modname)
            if '!' in stripped_modname else None)
        if chained:
            # ensure the stripped_modname is provided as by default the
            # handler will only deal with its own kind
            return chained.modname_source_to_target(
                toolchain, spec, stripped_modname, source)
        return stripped_modname

    def generate_handler_sourcepath(
            self, toolchain, spec, loaderplugin_sourcepath):
        """
        This returns the sourcepath (mapping) that may be added to the
        appropriate mapping within the spec for a successful toolchain
        execution.  The value generated is specific to the current
        spec that is being passed through the toolchain.

        The return value must be a modname: sourcepath mapping, e.g:

        return {
            'text': '/tmp/src/example_module/text/index.js',
            'json': '/tmp/src/example_module/json/index.js',
        }

        Subclasses must implement this to return a mapping of modnames
        the the absolute path of the desired sourcefiles.  Example:

        Implementation must also accept both the toolchain and the spec
        argument, along with the loaderplugin_sourcepath argument which
        will be a mapping of {modname: sourcepath} that are relevant to
        the current spec being processed through the toolchain.

        For nested/chained plugins, the recommended handling method is
        to make use of the assigned registry instance to lookup relevant
        loaderplugin handler(s) instances, and make use of their
        ``generate_handler_sourcepath`` method to generate the mapping
        required.  The immediate subclass in the loaderplugin module
        has a generic implementation done in this manner.
        """

        return {}

    def unwrap(self, value):
        """
        A helper method for unwrapping the loaderplugin fragment out of
        the provided value (typically a modname) and return it.

        Note that the filter chaining is very implementation specific to
        each and every loader plugin and their specific toolchain, so
        this default implementation is not going to attempt to consume
        everything in one go.

        Another note: if this is to be subclassed and if the return
        value does not actually remove the loaderplugin fragment, issues
        like default implmenetation of ``modname_source_to_target`` in
        this class to recurse forever.
        """

        globs = value.split('!', 1)
        if globs[0].split('?', 1)[0] == self.name:
            return globs[-1]
        else:
            return value

    def __call__(self, toolchain, spec, modname, source, target, modpath):
        """
        These need to provide the actual implementation required for the
        production of the final artifact, so this will need to locate
        the resources needed for this set of arguments to function.

        Implementations must return the associated modpaths, targets, and
        the export_module_name as a 3-tuple, after the copying or
        transpilation step was done.  Example:

        return (
            {'text!text_file.txt': 'text!/some/path/text_file.txt'},
            {'text_file.txt': 'text_file.txt'},
            ['text!text_file.txt'],
        )

        Note that implementations can trigger further lookups through
        the registry instance attached to this instance of the plugin,
        and implementations must also address the handling of this
        lookup and usage of the return values.

        Also note that while the toolchain and spec arguments are also
        provided, they should only be used for lookups; out of band
        modifications results in convoluted code flow.
        """

        raise NotImplementedError



# -*- coding: utf-8 -*-
import unittest

from os.path import join
from os import chdir
from os import makedirs
from pkg_resources import Distribution
from pkg_resources import Requirement
from pkg_resources import resource_filename
from pkg_resources import working_set as root_working_set

import calmjs.base
from calmjs.registry import Registry
from calmjs.registry import get as root_registry_get
from calmjs.registry import _inst as root_registry
from calmjs.base import BaseLoaderPluginHandler
from calmjs.loaderplugin import LoaderPluginRegistry
from calmjs.loaderplugin import LoaderPluginHandler
from calmjs.loaderplugin import NPMLoaderPluginHandler
from calmjs.loaderplugin import ModuleLoaderRegistry
from calmjs.module import ModuleRegistry
from calmjs.toolchain import NullToolchain
from calmjs.toolchain import Spec
from calmjs.utils import pretty_logging

from calmjs.testing.utils import remember_cwd
from calmjs.testing.utils import mkdtemp
from calmjs.testing.utils import stub_mod_working_set
from calmjs.testing.mocks import StringIO
from calmjs.testing.mocks import WorkingSet


class NotPlugin(LoaderPluginRegistry):
    """yeanah"""


class BadPlugin(LoaderPluginHandler):

    def __init__(self):
        """this will not be called; missing argument"""


class DupePlugin(LoaderPluginHandler):
    """
    Dummy duplicate plugin
    """


class LoaderPluginRegistryTestCase(unittest.TestCase):

    def test_to_plugin_name(self):
        registry = LoaderPluginRegistry(
            'calmjs.loader_plugin', _working_set=WorkingSet({}))
        self.assertEqual('example', registry.to_plugin_name('example'))
        self.assertEqual('example', registry.to_plugin_name('example?hi'))
        self.assertEqual('example', registry.to_plugin_name('example!hi'))
        self.assertEqual('example', registry.to_plugin_name('example?arg!hi'))

    def test_initialize_standard(self):
        # ensure that we have a proper working registry
        working_set = WorkingSet({'calmjs.loader_plugin': [
            'example = calmjs.loaderplugin:LoaderPluginHandler',
        ]})
        registry = LoaderPluginRegistry(
            'calmjs.loader_plugin', _working_set=working_set)
        plugin = registry.get('example')
        self.assertTrue(isinstance(plugin, LoaderPluginHandler))
        self.assertEqual({}, plugin.generate_handler_sourcepath(
            NullToolchain(), Spec(), {}))

    def test_initialize_failure_missing(self):
        working_set = WorkingSet({'calmjs.loader_plugin': [
            'not_plugin = calmjs.not_plugin:nothing',
        ]})
        # should not trigger import failure
        with pretty_logging(stream=StringIO()) as stream:
            registry = LoaderPluginRegistry(
                'calmjs.loader_plugin', _working_set=working_set)
        self.assertIsNone(registry.get('not_plugin'))
        self.assertIn(
            "registry 'calmjs.loader_plugin' failed to load loader plugin "
            "handler for entry point 'not_plugin =", stream.getvalue(),
        )

    def test_initialize_failure_not_plugin(self):
        working_set = WorkingSet({'calmjs.loader_plugin': [
            'not_plugin = calmjs.tests.test_loaderplugin:NotPlugin',
        ]})
        # should not trigger import failure
        with pretty_logging(stream=StringIO()) as stream:
            registry = LoaderPluginRegistry(
                'calmjs.loader_plugin', _working_set=working_set)
        self.assertIsNone(registry.get('not_plugin'))
        self.assertIn(
            "'not_plugin = calmjs.tests.test_loaderplugin:NotPlugin' does not "
            "lead to a valid loader plugin handler class",
            stream.getvalue()
        )

    def test_initialize_failure_bad_plugin(self):
        working_set = WorkingSet({'calmjs.loader_plugin': [
            'bad_plugin = calmjs.tests.test_loaderplugin:BadPlugin',
        ]}, dist=Distribution(project_name='plugin', version='1.0'))
        # should not trigger import failure
        with pretty_logging(stream=StringIO()) as stream:
            registry = LoaderPluginRegistry(
                'calmjs.loader_plugin', _working_set=working_set)
        self.assertIsNone(registry.get('bad_plugin'))
        self.assertIn(
            "registration of entry point "
            "'bad_plugin = calmjs.tests.test_loaderplugin:BadPlugin' from "
            "'plugin 1.0' to registry 'calmjs.loader_plugin' failed",
            stream.getvalue()
        )

    def test_initialize_warning_dupe_plugin(self):
        # ensure that we have a proper working registry
        working_set = WorkingSet({'calmjs.loader_plugin': [
            'example = calmjs.tests.test_loaderplugin:DupePlugin',
            'example = calmjs.loaderplugin:NPMLoaderPluginHandler',
        ]})
        # should not trigger import failure
        with pretty_logging(stream=StringIO()) as stream:
            registry = LoaderPluginRegistry(
                'calmjs.loader_plugin', _working_set=working_set)
        self.assertIn(
            "loader plugin handler for 'example' was already registered to an "
            "instance of 'calmjs.tests.test_loaderplugin:DupePlugin'",
            stream.getvalue()
        )
        # the second one will be registered
        self.assertTrue(
            isinstance(registry.get('example'), LoaderPluginHandler))
        # ensure that the handler can be acquired from a full name
        self.assertEqual('example', registry.get('example!hi').name)
        self.assertEqual('example', registry.get('example?arg!hi').name)
        self.assertEqual('example', registry.get('example?arg').name)
        self.assertIsNone(registry.get('examplearg'))
        self.assertIsNone(registry.get('ex'))
        self.assertIsNone(registry.get('ex!ample'))


class LoaderPluginHandlerTestcase(unittest.TestCase):

    def test_plugin_strip_basic(self):
        base = NPMLoaderPluginHandler(None, 'base')
        self.assertEqual(
            base.unwrap('base!some/dir/path.ext'),
            'some/dir/path.ext',
        )
        # unrelated will not be stripped.
        self.assertEqual(
            base.unwrap('something_else!some/dir/path.ext'),
            'something_else!some/dir/path.ext',
        )

    def test_plugin_unwrap_extras(self):
        base = LoaderPluginHandler(None, 'base')
        self.assertEqual(
            base.unwrap('base?someargument!some/dir/path.ext'),
            'some/dir/path.ext',
        )
        self.assertEqual(
            base.unwrap('base!other!some/dir/path.ext'),
            'other!some/dir/path.ext',
        )
        self.assertEqual(
            base.unwrap('base?arg!other?arg!some/dir/path.ext'),
            'other?arg!some/dir/path.ext',
        )

    def test_plugin_strip_edge(self):
        base = NPMLoaderPluginHandler(None, 'base')
        self.assertEqual(base.unwrap('base!'), '')

    def test_base_plugin_generate_handler_sourcepath(self):
        base = BaseLoaderPluginHandler(None, 'base')
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        self.assertEqual(
            base.generate_handler_sourcepath(toolchain, spec, {
                'base!bad': 'base!bad',
            }), {})

    def test_plugin_generate_handler_sourcepath_default_registry(self):
        base = LoaderPluginHandler(None, 'base')
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                base.generate_handler_sourcepath(toolchain, spec, {
                    'base!bad': 'base!bad',
                }), {})
        self.assertIn("using default loaderplugin registry", stream.getvalue())

    def test_plugin_generate_handler_sourcepath_resolved_registry(self):
        base = LoaderPluginHandler(None, 'base')
        reg = LoaderPluginRegistry('loaders', _working_set=WorkingSet({}))
        toolchain = NullToolchain()
        spec = Spec(
            working_dir=mkdtemp(self), calmjs_loaderplugin_registry=reg)
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                base.generate_handler_sourcepath(toolchain, spec, {
                    'base!bad': 'base!bad',
                }), {})
        self.assertIn(
            "loaderplugin registry 'loaders' already assigned to spec",
            stream.getvalue())

    def test_plugin_package_strip_broken_recursion_stop(self):
        class BadPluginHandler(LoaderPluginHandler):
            def unwrap(self, value):
                # return the identity
                return value

        base = BadPluginHandler(None, 'base')
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                base.generate_handler_sourcepath(toolchain, spec, {
                    'base!bad': 'base!bad',
                }), {})

        self.assertIn(
            "loaderplugin 'base' extracted same sourcepath of",
            stream.getvalue())

    def test_plugin_loaders_modname_source_to_target(self):
        class InterceptHandler(LoaderPluginHandler):
            def modname_source_to_target(self, *a, **kw):
                # failed to inspect and call parent
                return 'intercepted'

        reg = LoaderPluginRegistry('simloaders', _working_set=WorkingSet({}))
        base = reg.records['base'] = LoaderPluginHandler(reg, 'base')
        extra = reg.records['extra'] = LoaderPluginHandler(reg, 'extra')
        reg.records['intercept'] = InterceptHandler(reg, 'intercept')
        toolchain = NullToolchain()
        spec = Spec()
        self.assertEqual('fun.file', base.modname_source_to_target(
            toolchain, spec, 'base!fun.file', '/some/path/fun.file'))
        self.assertEqual('fun.file', extra.modname_source_to_target(
            toolchain, spec, 'extra!fun.file', '/some/path/fun.file'))
        self.assertEqual('fun.file', base.modname_source_to_target(
            toolchain, spec, 'extra!base!fun.file', '/some/path/fun.file'))
        # no plugin was found, so no modification
        self.assertEqual('noplugin!fun.file', base.modname_source_to_target(
            toolchain, spec, 'extra!noplugin!fun.file', '/some/path/fun.file'))
        # chained of the same type
        self.assertEqual('fun.file', base.modname_source_to_target(
            toolchain, spec, 'base!base!base!fun.file',
            '/some/path/fun.file'))
        # chained but overloaded
        self.assertEqual('intercepted', base.modname_source_to_target(
            toolchain, spec, 'base!intercept!base!fun.file',
            '/some/path/fun.file'))

    def test_plugin_loaders_modname_source_to_target_identity(self):
        # manually create a registry
        reg = LoaderPluginRegistry('simloaders', _working_set=WorkingSet({}))
        base = reg.records['local/dev'] = LoaderPluginHandler(reg, 'local/dev')
        toolchain = NullToolchain()
        spec = Spec()

        self.assertEqual('fun.file', base.modname_source_to_target(
            toolchain, spec, 'local/dev!fun.file',
            '/some/path/fun.file'))
        # a redundant usage test
        self.assertEqual('local/dev', base.modname_source_to_target(
            toolchain, spec, 'local/dev',
            '/some/path/to/the/plugin'))


class NPMPluginTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_plugin_base(self):
        base = NPMLoaderPluginHandler(None, 'base')
        with self.assertRaises(NotImplementedError):
            base(
                toolchain=None, spec=None, modname='', source='', target='',
                modpath='',
            )

    def test_plugin_package_base(self):
        base = NPMLoaderPluginHandler(None, 'base')
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                base.generate_handler_sourcepath(toolchain, spec, {}), {})
        self.assertIn(
            "no npm package name specified or could be resolved for "
            "loaderplugin 'base' of registry '<invalid_registry/handler>'; "
            "please subclass calmjs.loaderplugin:NPMLoaderPluginHandler such "
            "that the npm package name become specified", stream.getvalue(),
        )

    def test_plugin_package_missing_dir(self):
        base = NPMLoaderPluginHandler(None, 'base')
        base.node_module_pkg_name = 'dummy_pkg'
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                base.generate_handler_sourcepath(toolchain, spec, {}), {})
        self.assertIn(
            "could not locate 'package.json' for the npm package 'dummy_pkg' "
            "which was specified to contain the loader plugin 'base' in the "
            "current working directory '%s'" % spec['working_dir'],
            stream.getvalue(),
        )

    def test_plugin_package_missing_main(self):
        base = NPMLoaderPluginHandler(None, 'base')
        base.node_module_pkg_name = 'dummy_pkg'
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        pkg_dir = join(spec['working_dir'], 'node_modules', 'dummy_pkg')
        makedirs(pkg_dir)
        with open(join(pkg_dir, 'package.json'), 'w') as fd:
            fd.write('{}')

        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                base.generate_handler_sourcepath(toolchain, spec, {}), {})

        self.assertIn(
            "calmjs.loaderplugin 'package.json' for the npm package "
            "'dummy_pkg' does not contain a valid entry point: sources "
            "required for loader plugin 'base' cannot be included "
            "automatically; the build process may fail",
            stream.getvalue(),
        )

    def test_plugin_package_success_main(self):
        base = NPMLoaderPluginHandler(None, 'base')
        base.node_module_pkg_name = 'dummy_pkg'
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        pkg_dir = join(spec['working_dir'], 'node_modules', 'dummy_pkg')
        makedirs(pkg_dir)
        with open(join(pkg_dir, 'package.json'), 'w') as fd:
            fd.write('{"main": "base.js"}')

        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                join(pkg_dir, 'base.js'),
                base.generate_handler_sourcepath(toolchain, spec, {})['base'],
            )
        self.assertIn("for loader plugin 'base'", stream.getvalue())

    def test_plugin_package_success_package(self):
        base = NPMLoaderPluginHandler(None, 'base')
        base.node_module_pkg_name = 'dummy_pkg'
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        pkg_dir = join(spec['working_dir'], 'node_modules', 'dummy_pkg')
        makedirs(pkg_dir)
        with open(join(pkg_dir, 'package.json'), 'w') as fd:
            fd.write('{"browser": "browser/base.js"}')

        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                join(pkg_dir, 'browser', 'base.js'),
                base.generate_handler_sourcepath(toolchain, spec, {})['base'],
            )
        self.assertIn("for loader plugin 'base'", stream.getvalue())
        self.assertNotIn("missing working_dir", stream.getvalue())

    def test_plugin_package_success_package_spec_missing_working_dir(self):
        remember_cwd(self)
        cwd = mkdtemp(self)
        chdir(cwd)

        base = NPMLoaderPluginHandler(None, 'base')
        base.node_module_pkg_name = 'dummy_pkg'
        toolchain = NullToolchain()
        spec = Spec()
        pkg_dir = join(cwd, 'node_modules', 'dummy_pkg')
        makedirs(pkg_dir)
        with open(join(pkg_dir, 'package.json'), 'w') as fd:
            fd.write('{"browser": "browser/base.js"}')

        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                join(pkg_dir, 'browser', 'base.js'),
                base.generate_handler_sourcepath(toolchain, spec, {})['base'],
            )
        self.assertIn("for loader plugin 'base'", stream.getvalue())
        self.assertIn("missing working_dir", stream.getvalue())

    def test_plugin_package_dynamic_selection(self):

        class CustomHandler(NPMLoaderPluginHandler):
            def find_node_module_pkg_name(self, toolchain, spec):
                return spec.get('loaderplugin')

        reg = LoaderPluginRegistry('lp.reg', _working_set=WorkingSet({}))
        base = CustomHandler(reg, 'base')
        toolchain = NullToolchain()
        spec = Spec(working_dir=mkdtemp(self))
        pkg_dir = join(spec['working_dir'], 'node_modules', 'dummy_pkg')
        makedirs(pkg_dir)
        with open(join(pkg_dir, 'package.json'), 'w') as fd:
            fd.write('{"main": "base.js"}')

        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                {}, base.generate_handler_sourcepath(toolchain, spec, {}))
        self.assertIn(
            "no npm package name specified or could be resolved for "
            "loaderplugin 'base' of registry 'lp.reg'",
            stream.getvalue()
        )
        self.assertIn(
            "test_loaderplugin:CustomHandler may be at fault",
            stream.getvalue()
        )
        self.assertNotIn("for loader plugin 'base'", stream.getvalue())

        # plug the value into the spec to satisfy the condition for this
        # particular loader

        spec['loaderplugin'] = 'dummy_pkg'
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                join(pkg_dir, 'base.js'),
                base.generate_handler_sourcepath(toolchain, spec, {})['base'],
            )
        self.assertIn("base.js' for loader plugin 'base'", stream.getvalue())

    def create_base_extra_plugins(self, working_dir):
        # manually create a registry
        reg = LoaderPluginRegistry('simloaders', _working_set=WorkingSet({}))
        base = reg.records['base'] = NPMLoaderPluginHandler(reg, 'base')
        base.node_module_pkg_name = 'dummy_pkg1'
        extra = reg.records['extra'] = NPMLoaderPluginHandler(reg, 'extra')
        extra.node_module_pkg_name = 'dummy_pkg2'

        pkg_dir1 = join(working_dir, 'node_modules', 'dummy_pkg1')
        pkg_dir2 = join(working_dir, 'node_modules', 'dummy_pkg2')
        makedirs(pkg_dir1)
        makedirs(pkg_dir2)

        with open(join(pkg_dir1, 'package.json'), 'w') as fd:
            fd.write('{"main": "base.js"}')

        with open(join(pkg_dir2, 'package.json'), 'w') as fd:
            fd.write('{"main": "extra.js"}')
        return reg, base, extra, pkg_dir1, pkg_dir2

    def test_plugin_package_chained_loaders(self):
        working_dir = mkdtemp(self)
        reg, base, extra, base_dir, extra_dir = self.create_base_extra_plugins(
            working_dir)
        # standard case
        toolchain = NullToolchain()
        spec = Spec(working_dir=working_dir)
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                {'base': join(base_dir, 'base.js')},
                base.generate_handler_sourcepath(toolchain, spec, {
                    'base!fun.file': 'base!fun.file',
                }),
            )
        self.assertIn("for loader plugin 'base'", stream.getvalue())

        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual({
                'base': join(base_dir, 'base.js'),
                'extra': join(extra_dir, 'extra.js'),
            }, base.generate_handler_sourcepath(toolchain, spec, {
                    'base!fun.file': 'fun.file',
                    'base!extra!fun.file': 'fun.file',
                    'base!missing!fun.file': 'fun.file',
                    'base!extra!missing!fun.file': 'fun.file',
                }),
            )
        self.assertIn("for loader plugin 'base'", stream.getvalue())
        self.assertIn("for loader plugin 'extra'", stream.getvalue())
        self.assertNotIn("for loader plugin 'missing'", stream.getvalue())

        # for the outer one
        self.assertIn(
            "loaderplugin 'base' from registry 'simloaders' cannot find "
            "sibling loaderplugin handler for 'missing'; processing may fail "
            "for the following nested/chained sources: "
            "{'missing!fun.file': 'fun.file'}", stream.getvalue()
        )
        # for the inner one
        self.assertIn(
            "loaderplugin 'extra' from registry 'simloaders' cannot find "
            "sibling loaderplugin handler for 'missing'; processing may fail "
            "for the following nested/chained sources: "
            "{'missing!fun.file': 'fun.file'}", stream.getvalue()
        )

        # for repeat loaders
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual({
                'base': join(base_dir, 'base.js'),
                'extra': join(extra_dir, 'extra.js'),
            }, base.generate_handler_sourcepath(toolchain, spec, {
                    'base!extra!base!extra!fun.file': 'fun.file',
                }),
            )
        self.assertIn("for loader plugin 'base'", stream.getvalue())
        self.assertIn("for loader plugin 'extra'", stream.getvalue())

        # for repeat loaders
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual({
                'base': join(base_dir, 'base.js'),
            }, base.generate_handler_sourcepath(toolchain, spec, {
                    'base!base!base!fun.file': 'fun.file',
                }),
            )
        self.assertIn("for loader plugin 'base'", stream.getvalue())

        # for argument loaders
        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual({
                'base': join(base_dir, 'base.js'),
                'extra': join(extra_dir, 'extra.js'),
            }, base.generate_handler_sourcepath(toolchain, spec, {
                    'base?argument!extra?argument!fun.file': 'fun.file',
                }),
            )
        self.assertIn("for loader plugin 'base'", stream.getvalue())
        self.assertIn("for loader plugin 'extra'", stream.getvalue())

    def test_plugin_package_chained_loaders_initial_simple(self):
        working_dir = mkdtemp(self)
        reg, base, extra, base_dir, extra_dir = self.create_base_extra_plugins(
            working_dir)
        simple = reg.records['simple'] = LoaderPluginHandler(reg, 'simple')

        toolchain = NullToolchain()
        spec = Spec(working_dir=working_dir)

        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual(
                {},
                simple.generate_handler_sourcepath(toolchain, spec, {
                    'simple!fun.file': 'fun.file',
                }),
            )

        with pretty_logging(stream=StringIO()) as stream:
            self.assertEqual({
                'extra': join(extra_dir, 'extra.js'),
            }, simple.generate_handler_sourcepath(toolchain, spec, {
                    'simple!extra!fun.file': 'fun.file',
                }),
            )
        self.assertIn("for loader plugin 'extra'", stream.getvalue())

    def test_plugin_loaders_modname_source_to_target(self):
        working_dir = mkdtemp(self)
        reg, base, extra, base_dir, extra_dir = self.create_base_extra_plugins(
            working_dir)
        toolchain = NullToolchain()
        spec = Spec(working_dir=working_dir)
        self.assertEqual('fun.file', base.modname_source_to_target(
            toolchain, spec, 'base!fun.file', '/some/path/fun.file'))
        self.assertEqual('fun.file', base.modname_source_to_target(
            toolchain, spec, 'extra!base!fun.file', '/some/path/fun.file'))
        # no plugin was found, so no modification
        self.assertEqual('noplugin!fun.file', base.modname_source_to_target(
            toolchain, spec, 'extra!noplugin!fun.file', '/some/path/fun.file'))
        # chained of the same type
        self.assertEqual('fun.file', base.modname_source_to_target(
            toolchain, spec, 'base!base!base!fun.file',
            '/some/path/fun.file'))
        # a mismatched test


class ModuleLoaderRegistryTestCase(unittest.TestCase):
    """
    Test the module loader registry with a mocked working set with
    actual data to see resolution through the system.
    """

    def test_manual_construction_invalid_suffix_fail(self):
        with self.assertRaises(ValueError) as e:
            ModuleLoaderRegistry('some.module', _working_set=WorkingSet({}))
        self.assertEqual(
            "child module registry name defined with invalid suffix "
            "('some.module' does not end with '.loader')", str(e.exception))

    def test_manual_construction_parent_interactions(self):
        self.assertTrue(ModuleLoaderRegistry(
            'calmjs.module.loader', _working_set=WorkingSet({})))

        # forcibly stub out the real calmjs.module
        self.addCleanup(root_registry.records.pop, 'calmjs.module')
        root_registry.records['calmjs.module'] = None

        with self.assertRaises(ValueError) as e:
            ModuleLoaderRegistry(
                'calmjs.module.loader', _working_set=WorkingSet({}))
        self.assertEqual(
            "could not construct child module registry 'calmjs.module.loader' "
            "as its parent registry 'calmjs.module' could not be found",
            str(e.exception)
        )

    def test_module_loader_registry_integration(self):
        working_set = WorkingSet({
            'calmjs.module': [
                'module4 = calmjs.testing.module4',
            ],
            'calmjs.module.loader': [
                'css = css[style]',
            ],
            __name__: [
                'calmjs.module = calmjs.module:ModuleRegistry',
                'calmjs.module.loader = '
                'calmjs.loaderplugin:ModuleLoaderRegistry',
            ]},
            dist=Distribution(project_name='calmjs.testing', version='0.0')
        )
        stub_mod_working_set(self, [calmjs.base], working_set)

        # Not going to use the global registry, and using our custom
        # reservation entry
        local_root_registry = Registry(
            __name__, 'calmjs.testing', _working_set=working_set)

        with pretty_logging(stream=StringIO()):
            # silences "distribution 'calmjs.testing 0.0' not found"
            # warnings from stdout produced by the indexer, as the
            # provided working_set is invalid with entry points that do
            # not have a valid distribution.
            module_registry = root_registry_get('calmjs.module')
            module_loader_registry = root_registry_get('calmjs.module.loader')
            registry = local_root_registry.get_record('calmjs.module')
            loader_registry = local_root_registry.get_record(
                'calmjs.module.loader')

        self.assertIsNot(registry, module_registry)
        self.assertIsNot(loader_registry, module_loader_registry)
        self.assertEqual(
            sorted(k for k, v in registry.iter_records()), [
                'calmjs.testing.module4',
            ]
        )

        # test the basic items.
        results = registry.get_records_for_package('calmjs.testing')
        self.assertEqual(sorted(results.keys()), [
           'calmjs/testing/module4/widget',
        ])

        module4 = registry.get_record('calmjs.testing.module4')
        self.assertIn('calmjs/testing/module4/widget', module4)

        self.assertEqual({
            'css!calmjs/testing/module4/widget.style': resource_filename(
                'calmjs.testing', join('module4', 'widget.style')),
        }, loader_registry.get_records_for_package('calmjs.testing'))

        self.assertEqual(
            ['css'],
            loader_registry.get_loaders_for_package('calmjs.testing')
        )

    def test_module_loader_registry_multiple_loaders(self):
        working_set = WorkingSet({
            'calmjs.module': [
                'module4 = calmjs.testing.module4',
            ],
            'calmjs.module.loader': [
                'css = css[style,css]',
                'json = json[json]',
                'empty = empty[]',
            ],
            __name__: [
                'calmjs.module = calmjs.module:ModuleRegistry',
                'calmjs.module.loader = '
                'calmjs.loaderplugin:ModuleLoaderRegistry',
            ]},
            # use a real distribution instead for this case
            dist=root_working_set.find(Requirement.parse('calmjs')),
        )

        registry = ModuleRegistry('calmjs.module', _working_set=working_set)
        loader_registry = ModuleLoaderRegistry(
            'calmjs.module.loader', _working_set=working_set, _parent=registry)
        self.assertEqual({
            'calmjs': ['calmjs.testing.module4'],
        }, loader_registry.package_module_map)

        self.assertEqual(
            ['css', 'empty', 'json'],
            sorted(loader_registry.get_loaders_for_package('calmjs'))
        )

        self.assertEqual([
            'css!calmjs/testing/module4/other.css',
            'css!calmjs/testing/module4/widget.style',
            'json!calmjs/testing/module4/data.json',
        ], sorted(loader_registry.get_records_for_package('calmjs').keys()))

        # was not registered to calmjs.testing
        self.assertEqual([], loader_registry.get_loaders_for_package(
            'calmjs.testing'))
        self.assertEqual({}, loader_registry.get_records_for_package(
            'calmjs.testing'))



# -*- coding: utf-8 -*-
"""
This provides a "toolchain" for the calmjs framework.

The toolchain module provides two classes, a ``Spec`` which instances of
act as the orchestration object for the ``Toolchain`` instances, where
the states of a single workflow through it is tracked.  The other being
the ``Toolchain`` class, which `compiles`, `assembles` and `links` the
stuff as specified in the ``Spec`` into a standalone `module`, which can
be treated as an artifact file.  This whole thing jumpstarts that
process.

This whole thing started simply because the author thought it is a lot
easier to deal with JavaScript when one treats that as a compilation
target.  However given that the generic method works better, this module
encapsulates the generalized implementation of the previous version.

To better convey the understanding of how this came to place, the
following terms and variable name, prefix and suffixes have been defined
to carry the following meanings that is applicable throughout the entire
calmjs system:

modname
    A JavaScript path for a given module import system that it might
    have.  The Python analogue is the name of a given import module.
    Using the default mapper, one might map a Python module with the
    name ``calmjs.toolchain`` to ``calmjs/toolchain``.  While relative
    modpaths (i.e. identifiers beginning with './') are supported by
    most JavaScript/Node.js based import systems, its usage from within
    calmjs framework is discouraged.

sourcepath
    An absolute path on the local filesystem to the source file for a
    given modpath.  These two values (modpath and sourcepath) serves as
    the foundational mapping from a JavaScript module name to its
    corresponding source file.  Previously, this was simply named
    'source', so certain arguments remain named like so.

targetpath
    A relative path to a build directory (build_dir) serving as the
    write target for whatever proccessing done by the toolchain
    implementation to the file provided at the associated sourcepath.
    Note that by default this is NOT normalized to the currently running
    OS platform at generation time, due to varying and potentially
    complex internal/external integration usages.

    The relative path version (again, from build_dir) is typically
    recorded by instances of ``Spec`` objects that have undergone a
    ``Toolchain`` run.  Previously, this was simply named 'target'.

modpath
    Typically identical to modname, however the slight difference is
    that this is the transformed value that is to be better understood
    by the underlying tools.  Think of this as the post compiled value,
    or an alternative import location that is only applicable in the
    post-compiled context, specific to the toolchain class that it
    intends to encapsulate.
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import codecs
import errno
import logging
import re
import shutil
import sys
import warnings
from collections import namedtuple
from functools import partial
from inspect import currentframe
from traceback import format_stack
from os import mkdir
from os import makedirs
from os.path import basename
from os.path import join
from os.path import dirname
from os.path import exists
from os.path import isfile
from os.path import isdir
from os.path import normpath
from os.path import realpath
from tempfile import mkdtemp

from pkg_resources import Requirement
from pkg_resources import working_set as default_working_set

from calmjs.parse.io import read
from calmjs.parse.io import write
from calmjs.parse.parsers.es5 import parse
from calmjs.parse.unparsers.base import BaseUnparser
from calmjs.parse.unparsers.es5 import pretty_printer
from calmjs.parse.sourcemap import encode_sourcemap

from calmjs.base import BaseDriver
from calmjs.base import BaseRegistry
from calmjs.base import BaseLoaderPluginRegistry
from calmjs.base import PackageKeyMapping
from calmjs.registry import get as get_registry
from calmjs.exc import AdviceAbort
from calmjs.exc import AdviceCancel
from calmjs.exc import ValueSkip
from calmjs.exc import ToolchainAbort
from calmjs.exc import ToolchainCancel
from calmjs.utils import raise_os_error
from calmjs.utils import pdb_set_trace
from calmjs.vlqsm import SourceWriter

logger = logging.getLogger(__name__)

__all__ = [
    'AdviceRegistry', 'Spec', 'Toolchain', 'null_transpiler',

    'dict_setget', 'dict_setget_dict', 'dict_update_overwrite_check',

    'spec_update_loaderplugin_registry',
    'spec_update_sourcepath_filter_loaderplugins',

    'toolchain_spec_prepare_loaderplugins',

    'toolchain_spec_compile_entries', 'ToolchainSpecCompileEntry',

    'CALMJS_TOOLCHAIN_ADVICE',

    'SETUP', 'CLEANUP', 'SUCCESS',

    'AFTER_FINALIZE', 'BEFORE_FINALIZE', 'AFTER_LINK', 'BEFORE_LINK',
    'AFTER_ASSEMBLE', 'BEFORE_ASSEMBLE', 'AFTER_COMPILE', 'BEFORE_COMPILE',
    'AFTER_PREPARE', 'BEFORE_PREPARE', 'AFTER_TEST', 'BEFORE_TEST',

    'ADVICE_PACKAGES', 'ARTIFACT_PATHS', 'BUILD_DIR',
    'CALMJS_MODULE_REGISTRY_NAMES',
    'CALMJS_LOADERPLUGIN_REGISTRY_NAME',
    'CALMJS_LOADERPLUGIN_REGISTRY',
    'CALMJS_TEST_REGISTRY_NAMES',
    'CONFIG_JS_FILES', 'DEBUG',
    'EXPORT_MODULE_NAMES', 'EXPORT_PACKAGE_NAMES',
    'EXPORT_TARGET', 'EXPORT_TARGET_OVERWRITE',
    'SOURCE_MODULE_NAMES', 'SOURCE_PACKAGE_NAMES',
    'TEST_MODULE_NAMES', 'TEST_MODULE_PATHS_MAP', 'TEST_PACKAGE_NAMES',
    'TOOLCHAIN_BIN_PATH',
    'WORKING_DIR',
]

# these are the only non-key entities that should be in this module, as
# they currently reference auxilary registry classes that are currently
# residing in this module.
CALMJS_TOOLCHAIN_ADVICE = 'calmjs.toolchain.advice'
CALMJS_TOOLCHAIN_ADVICE_APPLY_SUFFIX = '.apply'

# define these as reserved advice names
SETUP = 'setup'
CLEANUP = 'cleanup'
SUCCESS = 'success'
AFTER_TEST = 'after_test'  # reserved however unused in this module
BEFORE_TEST = 'before_test'  # reserved however unused in this module
AFTER_FINALIZE = 'after_finalize'
BEFORE_FINALIZE = 'before_finalize'
AFTER_LINK = 'after_link'
BEFORE_LINK = 'before_link'
AFTER_ASSEMBLE = 'after_assemble'
BEFORE_ASSEMBLE = 'before_assemble'
AFTER_COMPILE = 'after_compile'
BEFORE_COMPILE = 'before_compile'
AFTER_PREPARE = 'after_prepare'
BEFORE_PREPARE = 'before_prepare'

# define these as reserved spec keys

# packages that have extra _optional_ advices supplied that have to be
# manually included.
ADVICE_PACKAGES = 'advice_packages'
# advice packages that have been applied to the spec via advice registry
# apply_toolchain_spec method.
ADVICE_PACKAGES_APPLIED_REQUIREMENTS = 'advice_packages_applied_requirements'
# listing of absolute locations on the file system where these bundled
# artifact files are.
ARTIFACT_PATHS = 'artifact_paths'
# build directory
BUILD_DIR = 'build_dir'
# the key for overriding the advice registry to be use
CALMJS_TOOLCHAIN_ADVICE_REGISTRY = 'calmjs_toolchain_advice_registry'
# source registries that have been used
CALMJS_MODULE_REGISTRY_NAMES = 'calmjs_module_registry_names'
CALMJS_TEST_REGISTRY_NAMES = 'calmjs_test_registry_names'
# loaderplugin registry related.
CALMJS_LOADERPLUGIN_REGISTRY_NAME = 'calmjs_loaderplugin_registry_name'
CALMJS_LOADERPLUGIN_REGISTRY = 'calmjs_loaderplugin_registry'
# configuration file for enabling execution of code in build directory
CONFIG_JS_FILES = 'config_js_files'
# for debug level
DEBUG = 'debug'
# the module names that have been exported out
EXPORT_MODULE_NAMES = 'export_module_names'
# the package names that have been exported out; not currently supported
# by any part of the library, but reserved nonetheless.
EXPORT_PACKAGE_NAMES = 'export_package_names'
# the container for the export target; either a file or directory; this
# should not be changed after the prepare step.
EXPORT_TARGET = 'export_target'
# specify that export target is safe to be overwritten.
EXPORT_TARGET_OVERWRITE = 'export_target_overwrite'
# for loaderplugin sourcepath dicts, where the maps are grouped by the
# name of the plugin; an intermediate step for processing of loader
# plugins
LOADERPLUGIN_SOURCEPATH_MAPS = 'loaderplugin_sourcepath_maps'
# if true, generate source map
GENERATE_SOURCE_MAP = 'generate_source_map'
# source module names; currently not supported by any part of the
# library, but reserved nonetheless
SOURCE_MODULE_NAMES = 'source_module_names'
# source Python package names that have been specified for source file
# extraction before the workflow, or the top level package(s); should
# not include automatically derived required packages.
SOURCE_PACKAGE_NAMES = 'source_package_names'
# for testing
# name of test modules
TEST_MODULE_NAMES = 'test_module_names'
# mapping of test module to their paths; i.e. sourcepath_map, but not
# labled as one to prevent naming conflicts (and they ARE to be
# standalone modules to be used directly by the toolchain and testing
# integration layer.
TEST_MODULE_PATHS_MAP = 'test_module_paths_map'
# name of test package
TEST_PACKAGE_NAMES = 'test_package_names'
# the binary that the toolchain encapsulates.
TOOLCHAIN_BIN_PATH = 'toolchain_bin_path'
# the working directory
WORKING_DIR = 'working_dir'


def cls_to_name(cls):
    return '%s:%s' % (cls.__module__, cls.__name__)


def _opener(*a):
    return codecs.open(*a, encoding='utf-8')


def partial_open(*a):
    return partial(codecs.open, *a, encoding='utf-8')


def _check_key_exists(spec, keys):
    for key in keys:
        if key not in spec:
            continue
        logger.error(
            "attempted to write '%s' to spec but key already exists; "
            "not overwriting, skipping", key
        )
        return True
    return False


def _deprecation_warning(msg):
    warnings.warn(msg, DeprecationWarning)
    logger.warning(msg)


def dict_setget(d, key, value):
    value = d[key] = d.get(key, value)
    return value


def dict_setget_dict(d, key):
    return dict_setget(d, key, {})


def dict_update_overwrite_check(base, fresh):
    """
    For updating a base dict with a fresh one, returning a list of
    3-tuples containing the key, previous value (base[key]) and the
    fresh value (fresh[key]) for all colliding changes (reassignment of
    identical values are omitted).
    """

    result = [
        (key, base[key], fresh[key])
        for key in set(base.keys()) & set(fresh.keys())
        if base[key] != fresh[key]
    ]
    base.update(fresh)
    return result


def log_exc_reason(
        etype, value, tb,
        msg="{etype} raised at {filename}:{lineno}; reason: {value}",
        log=logger.debug):
    while tb.tb_next is not None:
        tb = tb.tb_next
    frame = tb.tb_frame
    log(msg.format(
        etype=etype.__name__,
        filename=frame.f_code.co_filename,
        lineno=frame.f_lineno,
        value=value
    ))


# Spec functions for interfacing with loaderplugins
#
# The following functions (named in the format spec_*_loaderplugin_*)
# are helpers for extracting and filtering the mappings for interfacing
# with the loaderplugin helpers through the registry system.  These
# helpers are here as they are part of the toolchain, not as part of the
# loaderplugin module due to that module being the part that couples
# tightly with npm.

def spec_update_loaderplugin_registry(spec, default=None):
    """
    Resolve a BasePluginLoaderRegistry instance from spec, and update
    spec[CALMJS_LOADERPLUGIN_REGISTRY] with that value before returning
    it.
    """

    registry = spec.get(CALMJS_LOADERPLUGIN_REGISTRY)
    if isinstance(registry, BaseLoaderPluginRegistry):
        logger.debug(
            "loaderplugin registry '%s' already assigned to spec",
            registry.registry_name)
        return registry
    elif not registry:
        # resolving registry
        registry = get_registry(spec.get(CALMJS_LOADERPLUGIN_REGISTRY_NAME))
        if isinstance(registry, BaseLoaderPluginRegistry):
            logger.info(
                "using loaderplugin registry '%s'", registry.registry_name)
            spec[CALMJS_LOADERPLUGIN_REGISTRY] = registry
            return registry

    # acquire the real default instance, if possible.
    if not isinstance(default, BaseLoaderPluginRegistry):
        default = get_registry(default)
        if not isinstance(default, BaseLoaderPluginRegistry):
            logger.info(
                "provided default is not a valid loaderplugin registry")
            default = None

    if default is None:
        default = BaseLoaderPluginRegistry('<default_loaderplugins>')

    # TODO determine the best way to optionally warn about this for
    # toolchains that require this.
    if registry:
        logger.info(
            "object referenced in spec is not a valid loaderplugin registry; "
            "using default loaderplugin registry '%s'", default.registry_name)
    else:
        logger.info(
            "no loaderplugin registry referenced in spec; "
            "using default loaderplugin registry '%s'", default.registry_name)
    spec[CALMJS_LOADERPLUGIN_REGISTRY] = registry = default

    return registry


def spec_update_sourcepath_filter_loaderplugins(
        spec, sourcepath_map, sourcepath_map_key,
        loaderplugin_sourcepath_map_key=LOADERPLUGIN_SOURCEPATH_MAPS):
    """
    Take an existing spec and a sourcepath mapping (that could be
    produced via calmjs.dist.*_module_registry_dependencies functions)
    and split out the keys that does not contain loaderplugin syntax and
    assign it to the spec under sourcepath_key.

    For the parts with loader plugin syntax (i.e. modnames (keys) that
    contain a '!' character), they are instead stored under a different
    mapping under its own mapping identified by the plugin_name.  The
    mapping under loaderplugin_sourcepath_map_key will contain all
    mappings of this type.

    The resolution for the handlers will be done through the loader
    plugin registry provided via spec[CALMJS_LOADERPLUGIN_REGISTRY] if
    available, otherwise the registry instance will be acquired through
    the main registry using spec[CALMJS_LOADERPLUGIN_REGISTRY_NAME].

    For the example sourcepath_map input:

    sourcepath = {
        'module': 'something',
        'plugin!inner': 'inner',
        'plugin!other': 'other',
        'plugin?query!question': 'question',
        'plugin!plugin2!target': 'target',
    }

    The following will be stored under the following keys in spec:

    spec[sourcepath_key] = {
        'module': 'something',
    }

    spec[loaderplugin_sourcepath_map_key] = {
        'plugin': {
            'plugin!inner': 'inner',
            'plugin!other': 'other',
            'plugin?query!question': 'question',
            'plugin!plugin2!target': 'target',
        },
    }

    The goal of this function is to aid in processing each of the plugin
    types by batch, one level at a time.  It is up to the handler itself
    to trigger further lookups as there are implementations of loader
    plugins that do not respect the chaining mechanism, thus a generic
    lookup done at once may not be suitable.

    Note that nested/chained loaderplugins are not immediately grouped
    as they must be individually handled given that the internal syntax
    are generally proprietary to the outer plugin.  The handling will be
    dealt with at the Toolchain.compile_loaderplugin_entry method
    through the associated handler call method.

    Toolchain implementations may either invoke this directly as part
    of the prepare step on the required sourcepaths values stored in the
    spec, or implement this at a higher level before invocating the
    toolchain instance with the spec.
    """

    default = dict_setget_dict(spec, sourcepath_map_key)
    registry = spec_update_loaderplugin_registry(spec)

    # it is more loaderplugin_sourcepath_maps
    plugins = dict_setget_dict(spec, loaderplugin_sourcepath_map_key)

    for modname, sourcepath in sourcepath_map.items():
        parts = modname.split('!', 1)
        if len(parts) == 1:
            # default
            default[modname] = sourcepath
            continue

        # don't actually do any processing yet.
        plugin_name = registry.to_plugin_name(modname)
        plugin = dict_setget_dict(plugins, plugin_name)
        plugin[modname] = sourcepath


def toolchain_spec_prepare_loaderplugins(
        toolchain, spec,
        loaderplugin_read_key,
        handler_sourcepath_key,
        loaderplugin_sourcepath_map_key=LOADERPLUGIN_SOURCEPATH_MAPS):
    """
    A standard helper function for combining the filtered (e.g. using
    ``spec_update_sourcepath_filter_loaderplugins``) loaderplugin
    sourcepath mappings back into one that is usable with the standard
    ``toolchain_spec_compile_entries`` function.

    Arguments:

    toolchain
        The toolchain
    spec
        The spec

    loaderplugin_read_key
        The read_key associated with the loaderplugin process as set up
        for the Toolchain that implemented this.  If the toolchain has
        this in its compile_entries:

            ToolchainSpecCompileEntry('loaderplugin', 'plugsrc', 'plugsink')

        The loaderplugin_read_key it must use will be 'plugsrc'.

    handler_sourcepath_key
        All found handlers will have their handler_sourcepath method be
        invoked, and the combined results will be a dict stored in the
        spec under that key.

    loaderplugin_sourcepath_map_key
        It must be the same key to the value produced by
        ``spec_update_sourcepath_filter_loaderplugins``
    """

    # ensure the registry is applied to the spec
    registry = spec_update_loaderplugin_registry(
        spec, default=toolchain.loaderplugin_registry)

    # this one is named like so for the compile entry method
    plugin_sourcepath = dict_setget_dict(
        spec, loaderplugin_read_key + '_sourcepath')
    # the key is supplied by the toolchain that might make use of this
    if handler_sourcepath_key:
        handler_sourcepath = dict_setget_dict(spec, handler_sourcepath_key)
    else:
        # provide a null value for this.
        handler_sourcepath = {}

    for key, value in spec.get(loaderplugin_sourcepath_map_key, {}).items():
        handler = registry.get(key)
        if handler:
            # assume handler will do the job.
            logger.debug("found handler for '%s' loader plugin", key)
            plugin_sourcepath.update(value)
            logger.debug(
                "plugin_sourcepath updated with %d keys", len(value))
            # TODO figure out how to address the case where the actual
            # JavaScript module for the handling wasn't found.
            handler_sourcepath.update(
                handler.generate_handler_sourcepath(toolchain, spec, value))
        else:
            logger.warning(
                "loaderplugin handler for '%s' not found in loaderplugin "
                "registry '%s'; as arguments associated with loader plugins "
                "are specific, processing is disabled for this group; the "
                "sources referenced by the following names will not be "
                "compiled into the build target: %s",
                key, registry.registry_name, sorted(value.keys()),
            )


def toolchain_spec_compile_entries(
        toolchain, spec, entries, process_name, overwrite_log=None):
    """
    The standardized Toolchain Spec Entries compile function

    This function accepts a toolchain instance, the spec to be operated
    with and the entries provided for the process name.  The standard
    flow is to deferr the actual processing to the toolchain method
    `compile_{process_name}_entry` for each entry in the entries list.

    The generic compile entries method for the compile process.

    Arguments:

    toolchain
        The toolchain to be used for the operation.
    spec
        The spec to be operated with.
    entries
        The entries for the source.
    process_name
        The name of the specific compile process of the provided
        toolchain.
    overwrite_log
        A callable that will accept a 4-tuple of suffix, key, original
        and new value, if monitoring of overwritten values are required.
        suffix is derived from the modpath_suffix or targetpath_suffix
        of the toolchain instance, key is the key on any of the keys on
        either of those mappings, original and new are the original and
        the replacement value.
    """

    processor = getattr(toolchain, 'compile_%s_entry' % process_name)
    modpath_logger = (
        partial(overwrite_log, toolchain.modpath_suffix)
        if callable(overwrite_log) else None)
    targetpath_logger = (
        partial(overwrite_log, toolchain.targetpath_suffix)
        if callable(overwrite_log) else None)
    return process_compile_entries(
        processor, spec, entries, modpath_logger, targetpath_logger)


def process_compile_entries(
        processor, spec, entries, modpath_logger=None, targetpath_logger=None):
    """
    The generalized raw spec entry process invocation loop.
    """

    # Contains a mapping of the module name to the compiled file's
    # relative path starting from the base build_dir.
    all_modpaths = {}
    all_targets = {}
    # List of exported module names, should be equal to all keys of
    # the compiled and bundled sources.
    all_export_module_names = []

    def update(base, fresh, logger):
        if callable(logger):
            for dupes in dict_update_overwrite_check(base, fresh):
                logger(*dupes)
        else:
            base.update(fresh)

    for entry in entries:
        modpaths, targetpaths, export_module_names = processor(spec, entry)
        update(all_modpaths, modpaths, modpath_logger)
        update(all_targets, targetpaths, targetpath_logger)
        all_export_module_names.extend(export_module_names)

    return all_modpaths, all_targets, all_export_module_names


ToolchainSpecCompileEntry = namedtuple('ToolchainSpecCompileEntry', [
    'process_name', 'read_key', 'store_key', 'logger', 'log_level'])
ToolchainSpecCompileEntry.__new__.__defaults__ = (None, None)


def debugger(spec, extras):
    if not spec.get(DEBUG):
        return
    for key in extras:
        if not key.startswith('debug_'):
            continue
        name = key.split('_', 1)[1]
        logger.debug("debugger advised at '%s'", name)
        spec.advise(name, pdb_set_trace)


def null_transpiler(spec, reader, writer):
    writer.write(reader.read())


class Spec(dict):
    """
    Instances of these will track the progress through a Toolchain
    instance.
    """

    def __init__(self, *a, **kw):
        self._deprecation_match_4_0 = [(re.compile(p), r) for p, r in (
            ('^((?!generate)(.*))_source_map$', '\\1_sourcepath'),
            ('_targets$', '_targetpaths'),
        )]

        clean_kw = {
            self.__process_deprecated_key(k): v for k, v in kw.items()}

        super(Spec, self).__init__(*a, **clean_kw)
        self._advices = {}
        self._frames = {}
        self._called = set()

    def __process_deprecated_key(self, key):
        for patt, repl in self._deprecation_match_4_0:
            if patt.search(key):
                break
        else:
            return key

        new_key = patt.sub(repl, key)
        _deprecation_warning(
            "Spec key '%s' has been remapped to '%s' in calmjs-3.0.0; this "
            "automatic remap will be removed by calmjs-4.0.0" % (key, new_key)
        )
        return new_key

    def get(self, key, default=NotImplemented):
        key = self.__process_deprecated_key(key)
        if default is NotImplemented:
            return dict.get(self, key)
        else:
            return dict.get(self, key, default)

    def __getitem__(self, key):
        return dict.__getitem__(
            self, self.__process_deprecated_key(key))

    def __setitem__(self, key, value):
        return dict.__setitem__(
            self, self.__process_deprecated_key(key), value)

    def __repr__(self):
        debug = self.get(DEBUG)
        if not isinstance(debug, int) or debug < 2:
            return object.__repr__(self)
        # for a repr, helpers by pprint module and even json doesn't
        # deal with circular references for this, so just use the trusty
        # parent class and be done with it (even though I wanted a
        # sorted output, this is fine for now as I don't want to spam
        # logs without debugging enabled).
        return dict.__repr__(self)

    def update_selected(self, other, selected):
        """
        Like update, however a list of selected keys must be provided.
        """

        self.update({k: other[k] for k in selected})

    def __advice_stack_frame_protection(self, frame):
        """
        Overriding of this is only permitted if and only if your name is
        Megumin and you have a pet/familiar named Chomusuke.
        """

        if frame is None:
            logger.debug(
                'currentframe() returned None; frame protection disabled')
            return

        f_back = frame.f_back
        while f_back:
            if f_back.f_code is self.handle.__code__:
                raise RuntimeError(
                    "indirect invocation of '%s' by 'handle' is forbidden" %
                    frame.f_code.co_name,
                )
            f_back = f_back.f_back

    def advise(self, name, f, *a, **kw):
        """
        Add an advice that will be handled later by the handle method.

        Arguments:

        name
            The name of the advice group
        f
            A callable method or function.

        The rest of the arguments will be passed as arguments and
        keyword arguments to f when it's invoked.
        """

        if name is None:
            return

        advice = (f, a, kw)
        debug = self.get(DEBUG)

        frame = currentframe()
        if frame is None:
            logger.debug('currentframe() failed to return frame')
        else:
            if name in self._called:
                self.__advice_stack_frame_protection(frame)
            if debug:
                logger.debug(
                    "advise '%s' invoked by %s:%d",
                    name,
                    frame.f_back.f_code.co_filename, frame.f_back.f_lineno,
                )
                if debug > 1:
                    # use the memory address of the tuple which should
                    # be stable
                    self._frames[id(advice)] = ''.join(
                        format_stack(frame.f_back))

        self._advices[name] = self._advices.get(name, [])
        self._advices[name].append(advice)

    def handle(self, name):
        """
        Call all advices at the provided name.

        This has an analogue in the join point in aspected oriented
        programming, but the analogy is a weak one as we don't have the
        proper metaobject protocol to support this.  Implementation that
        make use of this system should make it clear that they will call
        this method with name associated with its group before and after
        its execution, or that the method at hand that want this invoked
        be called by this other conductor method.

        For the Toolchain standard steps (prepare, compile, assemble,
        link and finalize), this handle method will only be called by
        invoking the toolchain as a callable.  Calling those methods
        piecemeal will not trigger the invocation, even though it
        probably should.  Modules, classes and methods that desire to
        call their own handler should instead follow the convention
        where the handle be called before and after with the appropriate
        names.  For instance:

            def test(self, spec):
                spec.handle(BEFORE_TEST)
                # do the things
                spec.handle(AFTER_TEST)

        This arrangement will need to be revisited when a proper system
        is written at the metaclass level.

        Arguments:

        name
            The name of the advices group.  All the callables
            registered to this group will be invoked, last-in-first-out.
        """

        if name in self._called:
            logger.warning(
                "advice group '%s' has been called for this spec %r",
                name, self,
            )
            # only now ensure checking
            self.__advice_stack_frame_protection(currentframe())
        else:
            self._called.add(name)

        # Get a complete clone, so indirect manipulation done to the
        # reference that others have access to will not have an effect
        # within the scope of this execution.  Please refer to the
        # test_toolchain, test_spec_advice_no_infinite_pop test case.
        advices = []
        advices.extend(self._advices.get(name, []))

        if advices and self.get('debug'):
            logger.debug(
                "handling %d advices in group '%s' ", len(advices), name)

        while advices:
            try:
                # advice processing is done lifo (last in first out)
                values = advices.pop()
                advice, a, kw = values
                if not ((callable(advice)) and
                        isinstance(a, tuple) and
                        isinstance(kw, dict)):
                    raise TypeError
            except ValueError:
                logger.info('Spec advice extraction error: got %s', values)
            except TypeError:
                logger.info('Spec advice malformed: got %s', values)
            else:
                try:
                    try:
                        advice(*a, **kw)
                    except Exception as e:
                        # get that back by the id.
                        frame = self._frames.get(id(values))
                        if frame:
                            logger.info('Spec advice exception: %r', e)
                            logger.info(
                                'Traceback for original advice:\n%s', frame)
                        # continue on for the normal exception
                        raise
                except AdviceCancel as e:
                    logger.info(
                        "advice %s in group '%s' signaled its cancellation "
                        "during its execution: %s", advice, name, e
                    )
                    if self.get(DEBUG):
                        logger.debug(
                            'showing traceback for cancellation', exc_info=1,
                        )
                except AdviceAbort as e:
                    # this is a signaled error with a planned abortion
                    logger.warning(
                        "advice %s in group '%s' encountered a known error "
                        "during its execution: %s; continuing with toolchain "
                        "execution", advice, name, e
                    )
                    if self.get(DEBUG):
                        logger.warning(
                            'showing traceback for error', exc_info=1,
                        )
                except ToolchainCancel:
                    # this is the safe cancel
                    raise
                except ToolchainAbort as e:
                    logger.critical(
                        "an advice in group '%s' triggered an abort: %s",
                        name, str(e)
                    )
                    raise
                except KeyboardInterrupt:
                    raise ToolchainCancel('interrupted')
                except Exception as e:
                    # a completely unplanned failure
                    logger.critical(
                        "advice %s in group '%s' terminated due to an "
                        "unexpected exception: %s", advice, name, e
                    )
                    if self.get(DEBUG):
                        logger.critical(
                            'showing traceback for error', exc_info=1,
                        )


class AdviceRegistry(BaseRegistry):
    """
    Registry for Spec.advise application functions.

    The declaration is specific to one given toolchain, and they are
    declared as EntryPoints by packages.  Once defined, the package may
    be refernced as an Advice Package and it may be specified by the
    spec key ADVICE_PACKAGES.

    For example, if 'example.package' declares the following::

        [calmjs.toolchain.advice]
        example.package.toolchain:Toolchain = example.package.spec:apply

    And if that toolchain was invoked with a Spec that has the following
    definition:

        Spec({ADVICE_PACKAGES: ['example.package[extra1,extra2]']})

    Then the target specified by that entry_point will then be invoked
    with the spec and the extras passed as an unordered list.

    For the implementation to function as expected, it requires the
    Toolchain to invoke process_toolchain_spec_package of instances of
    this registry.
    """

    def _init(self):
        for entry_point in self.raw_entry_points:
            key = entry_point.dist.project_name
            records = self.records[key] = self.records.get(key, {})
            records[entry_point.name] = entry_point

    def _to_requirement(self, value):
        try:
            return Requirement.parse(value)
        except ValueError as e:
            logger.error(
                "the specified value '%s' for advice setup is not valid for "
                "a package/requirement: %s", value, e,
            )
            raise

    def get_record(self, name):
        return self.records.get(name)

    def applied_requirements_map_from_spec(self, toolchain, spec):
        # it may be good to warn about requirements that have been
        # replaced by the standalone method.
        return {
            req.name: req
            for req in spec.get(ADVICE_PACKAGES_APPLIED_REQUIREMENTS, [])
        }

    def apply_toolchain_spec(self, toolchain, spec):
        """
        Apply the advice packages as defined by this registry to the
        provided toolchain and spec.

        This implementation will first apply whatever ADVICE_PACKAGES
        are provided by the spec, before applying whatever else that may
        be applied by this registry instance.  Before application of the
        advice packages, the ADVICE_PACKAGES_APPLIED_REQUIREMENTS key
        from the spec will also be checked first to prevent the
        application of advice with the same requirement name.

        As the ADVICE_PACKAGES feature was originally implemented as a
        part of the SETUP advice applied by the Runtime class, and the
        implementation allowed multiple copies of the same requirement
        be applied, this feature will be maintained as this method
        implements a fully contained version of that along with the
        version that applies the ones recorded by the registry.
        However, multiple execution of this method will not reapply
        the ones that have been recorded as applied.
        """

        # first step: apply all the advice packages as found in the
        # provided spec, as these are specified to be necessary which
        # may override whatever other requirements might be specified
        # in the accompanied apply registry.
        spec_advice_packages = spec.get(ADVICE_PACKAGES) or []
        # construct a mapping based on the list of applied requirements
        # that have been also recorded on this spec by the common apply
        # standalone method.
        applied_req_map = self.applied_requirements_map_from_spec(
            toolchain, spec)
        newly_applied_req_map = {}

        logger.debug(
            "invoking apply_toolchain_spec using instance of %s named '%s'",
            cls_to_name(type(self)), self.registry_name,
        )

        for value in spec_advice_packages:
            try:
                req = self._to_requirement(value)
            except ValueError:
                # error log entry already generated by the above method.
                continue

            if req.name in applied_req_map:
                logger.warning(
                    "advice package '%s' already applied as '%s'; skipping",
                    req, applied_req_map[req.name]
                )
                continue

            if req.name in newly_applied_req_map:
                logger.warning(
                    "advice package '%s' was previously applied as '%s'; "
                    "the recommended usage manner is to only specify any "
                    "given advice package once complete with all the required "
                    "extras, and that underlying implementation be structured "
                    "in a manner that support this one-shot invocation "
                    "format",
                    req, newly_applied_req_map[req.name]
                )

            logger.debug("applying advice package '%s'", value)
            self._process_toolchain_spec_requirement(toolchain, spec, req)
            # still going to warn
            newly_applied_req_map[req.name] = req

        # finally, find the accompanied apply registry for the package
        # definitions
        advice_apply_registry_key = (
            self.registry_name + CALMJS_TOOLCHAIN_ADVICE_APPLY_SUFFIX)
        advice_apply_registry = get_registry(advice_apply_registry_key)

        if not isinstance(advice_apply_registry, AdviceApplyRegistry):
            logger.warning(
                "registry key '%s' resulted in %r which is not a valid advice "
                "apply registry; no package level advice apply steps will be "
                "applied", advice_apply_registry_key, advice_apply_registry
            )
            return

        # combine the newly applied ones with existing ones.
        applied_req_map.update(newly_applied_req_map)

        for pkg_name in spec.get(SOURCE_PACKAGE_NAMES, []):
            requirements = advice_apply_registry.get_record(pkg_name)
            if not requirements:
                continue
            logger.info(
                "source package '%s' specified %d advice package(s) to be "
                "applied", pkg_name, len(requirements)
            )

            for req in requirements:
                if req.name in applied_req_map:
                    logger.debug(
                        "skipping specified advice package '%s' as '%s' was "
                        "already applied", req, applied_req_map[req.name]
                    )
                    continue
                logger.debug("apply advice package '%s'", req)
                self._process_toolchain_spec_requirement(toolchain, spec, req)
                applied_req_map[req.name] = req

    def process_toolchain_spec_package(self, toolchain, spec, value):
        # the original one-shot method.
        try:
            req = self._to_requirement(value)
        except ValueError:
            pass
        else:
            return self._process_toolchain_spec_requirement(
                toolchain, spec, req)

    def _process_toolchain_spec_requirement(self, toolchain, spec, req):
        if not isinstance(toolchain, Toolchain):
            logger.debug(
                'apply_toolchain_spec or process_toolchain_spec_package '
                'must be invoked with a toolchain instance, not %s', toolchain,
            )
            return

        pkg_name = req.project_name
        toolchain_cls = type(toolchain)
        toolchain_advices = self.get_record(pkg_name)

        if toolchain_advices is None and not default_working_set.find(req):
            logger.warning(
                "advice setup steps required from package/requirement "
                "'%s', however it is not found or not installed in this "
                "environment; build may continue; if there are errors or "
                "unexpected behavior that occur, it may be corrected by "
                "providing the missing requirement into this environment "
                "by installing the relevant package", req
            )
            return
        elif not toolchain_advices:
            logger.debug(
                "no advice setup steps registered for package/requirement "
                "'%s'", req)
            return

        logger.debug(
            "found advice setup steps registered for package/requirement "
            "'%s'; checking for compatibility with toolchain '%s'",
            req, cls_to_name(toolchain_cls)
        )

        entry_points = [
            toolchain_advices.get(cls_name) for cls_name in (
                cls_to_name(cls) for cls in toolchain_cls.__mro__
                if issubclass(cls, Toolchain)
            ) if cls_name in toolchain_advices
        ]

        if not entry_points:
            logger.debug("no compatible advice setup steps found")

        for entry_point in entry_points:
            try:
                f = entry_point.load()
            except ImportError:
                logger.error(
                    "ImportError: entry_point '%s' in group '%s' while "
                    "processing toolchain spec advice setup step "
                    "registered under advice package '%s'",
                    entry_point, self.registry_name, req
                )
                continue

            try:
                f(spec, sorted(req.extras))
            except Exception:
                logger.exception(
                    "failure encountered while setting up advices through "
                    "entry_point '%s' in group '%s' "
                    "registered under advice package '%s'",
                    entry_point, self.registry_name, req
                )
            else:
                logger.debug(
                    "entry_point '%s' registered by advice package '%s' "
                    "applied as an advice setup step by %s '%s'",
                    entry_point, req,
                    cls_to_name(type(self)), self.registry_name,
                )

        # will just simply be applied regardless.
        spec.setdefault(ADVICE_PACKAGES_APPLIED_REQUIREMENTS, [])
        spec[ADVICE_PACKAGES_APPLIED_REQUIREMENTS].append(req)


class AdviceApplyRegistry(BaseRegistry):
    """
    Registry to automatically set up the the list of ADVICE_PACKAGES for
    a given Toolchain execution.  If a package has entries declared in
    this registry, the default Toolchain will apply those declarations
    onto the list of ADVICE_PACKAGES whenever it appears as a member of
    SOURCE_PACKAGE_NAMES.  For example (continuing on from the example
    in AdviceRegistry), if 'example.demo' declares the following:

        [calmjs.toolchain.advice.apply]
        example.demo = example.package[extra3,extra4]

    Whenever 'example.demo' appears as an entry in SOURCE_PACKAGE_NAMES
    in a spec, the following additional flags will be applied to the
    spec.

        {ADVICE_PACKAGES: ['example.package[extra1,extra2]']}

    Naturally, this would not be carried across dependents - if
    dependents also need that exact advice setup applied, it needs to
    also declare the same entry in this registry.  Also, in order for
    the implementation to function as expected, the standard Toolchain
    must be used, and the relevant functionality should be invoked and
    not be overridden.

    Note that the key is ignored under the current implementation.
    """

    def _init(self):
        # since the record keys are package names
        self.records = PackageKeyMapping()
        for entry_point in self.raw_entry_points:
            self._init_entry_point(entry_point)

    def _init_entry_point(self, entry_point):
        if not entry_point.dist:
            logger.warning(
                'entry_points passed to %s for registration must provide a '
                'distribution with a project name; registration of %s skipped',
                cls_to_name(type(self)), entry_point,
            )
            return
        key = entry_point.dist.project_name
        self.records.setdefault(key, [])
        # have to cast the entry point into
        try:
            requirement = Requirement.parse(str(entry_point).split('=', 1)[1])
        except ValueError as e:
            logger.warning(
                "entry_point '%s' cannot be registered to %s due to the "
                "following error: %s",
                entry_point, cls_to_name(type(self)), e
            )
        else:
            self.records[key].append(requirement)

    def get_record(self, name):
        return self.records.get(name)


class Toolchain(BaseDriver):
    """
    For shared methods between all toolchains.

    The objective of this class is to provide a consistent interface
    from calmjs to the various cli Node.js tools, this class inherits
    from the BaseDriver class.  This means having the same foundation
    and also the ability to reuse a number of useful utility methods
    for talking to those scripts and binaries.

    This also involves some standardized processes within the calmjs
    framework, naming the definition of the following items that every
    subclass and implementation must support.

    BUILD_DIR
        The build directory.  This can be manually specified, or be a
        temporary directory automatically created and destroyed.
    CALMJS_MODULE_REGISTRY_NAMES
        For recording the list of calmjs modules it has used, so other
        parts of the framework can make use of this, such as inferring
        which test modules it should use.
    EXPORT_TARGET
        A path on the filesystem that this toolchain will ultimately
        generate its output to.
    EXPORT_TARGET_OVERWRITE
        Signifies that the export target can be safely overwritten.
    WORKING_DIR
        The working directory where the relative paths will be based
        from.
    """

    # subclasses may assign an identifier or instance of a compatible
    # loaderplugin registry for use with the encapsulated framework.
    loaderplugin_registry = None

    def __init__(self, *a, **kw):
        """
        Refer to parent for exact arguments.
        """

        super(Toolchain, self).__init__(*a, **kw)
        self.opener = _opener
        self.setup_filename_suffix()
        self.setup_transpiler()
        self.setup_prefix_suffix()
        self.setup_compile_entries()

    # Helpers

    def realpath(self, spec, key):
        """
        Resolve and update the path key in the spec with its realpath,
        based on the working directory.
        """

        if key not in spec:
            # do nothing for now
            return

        if not spec[key]:
            logger.warning(
                "cannot resolve realpath of '%s' as it is not defined", key)
            return

        check = realpath(join(spec.get(WORKING_DIR, ''), spec[key]))
        if check != spec[key]:
            spec[key] = check
            logger.warning(
                "realpath of '%s' resolved to '%s', spec is updated",
                key, check
            )
        return check

    # Setup related methods

    def setup_filename_suffix(self):
        """
        Set up the filename suffix for the sources and targets.
        """

        self.filename_suffix = '.js'

    def setup_transpiler(self):
        """
        Subclasses will need to implement this to setup the transpiler
        attribute, which the compile method will invoke.
        """

        self.parser = NotImplemented
        self.transpiler = NotImplemented

    def setup_prefix_suffix(self):
        """
        Set up the compile prefix, sourcepath and the targetpath suffix
        attributes, which are the prefix to the function name and the
        suffixes to retrieve the values from for creating the generator
        function.
        """

        self.compile_prefix = 'compile_'
        self.sourcepath_suffix = '_sourcepath'
        self.modpath_suffix = '_modpaths'
        self.targetpath_suffix = '_targetpaths'

    # TODO BBB backward compat fixes

    @property
    def sourcemap_suffix(self):
        _deprecation_warning(
            'sourcemap_suffix has been renamed to sourcepath_suffix; '
            'Toolchain attribute will be removed by calmjs-4.0.0',
        )
        return self.sourcepath_suffix

    @sourcemap_suffix.setter
    def sourcemap_suffix(self, value):
        _deprecation_warning(
            'sourcemap_suffix has been renamed to sourcepath_suffix; '
            'Toolchain attribute will be removed by calmjs-4.0.0',
        )
        self.sourcepath_suffix = value

    @property
    def target_suffix(self):
        _deprecation_warning(
            'target_suffix has been renamed to targetpath_suffix; '
            'Toolchain attribute will be removed by calmjs-4.0.0'
        )
        return self.targetpath_suffix

    @target_suffix.setter
    def target_suffix(self, value):
        _deprecation_warning(
            'target_suffix has been renamed to targetpath_suffix; '
            'Toolchain attribute will be removed by calmjs-4.0.0'
        )
        self.targetpath_suffix = value

    def setup_compile_entries(self):
        """
        The method that sets up the map that maps the compile methods
        stored in this class instance to the spec key that the generated
        maps should be stored at.
        """

        self.compile_entries = self.build_compile_entries()

    def build_compile_entries(self):
        """
        Build the entries that will be used to acquire the methods for
        the compile step.

        This is to be a list of 3-tuples.

        first element being the method name, which is a name that will
        be prefixed with the compile_prefix, default being `compile_`;
        alternatively a callable could be provided.  This method must
        return a 2-tuple.

        second element being the read key for the sourcepath dict, which
        is the name to be read from the spec and it will be suffixed
        with the sourcepath_suffix, default being `_sourcepath`.

        third element being the write key for first return value of the
        method, it will be suffixed with the modpath_suffix, defaults to
        `_modpaths`, and the targetpath_suffix, default to `_targetpaths`.

        The method referenced SHOULD NOT assign values to the spec, and
        it must produce and return a 2-tuple:

        first element should be the map from the module to the written
        targets, the key being the module name (modname) and the value
        being the relative path of the final file to the build_dir

        the second element must be a list of module names that it
        exported.
        """

        return (
            # compile_*, *_sourcepath, (*_modpaths, *_targetpaths)
            ToolchainSpecCompileEntry('transpile', 'transpile', 'transpiled'),
            ToolchainSpecCompileEntry('bundle', 'bundle', 'bundled'),
        )

    # Default built-in methods referenced by methods that will be
    # executed, as constructed by build_compile_entries.

    # Following are used for the transpile and bundle compile processes.

    def _validate_build_target(self, spec, target):
        """
        Essentially validate that the target is inside the build_dir.
        """

        if not realpath(target).startswith(spec[BUILD_DIR]):
            raise ValueError('build_target %s is outside build_dir' % target)

    # note that in the following methods, a shorthand notation is used
    # for some of the arguments: nearly all occurrences of source means
    # sourcepath, and target means targetpath.

    def _generate_transpile_target(self, spec, target):
        # ensure that the target is fully normalized.
        bd_target = join(spec[BUILD_DIR], normpath(target))
        self._validate_build_target(spec, bd_target)
        if not exists(dirname(bd_target)):
            logger.debug("creating dir '%s'", dirname(bd_target))
            makedirs(dirname(bd_target))

        return bd_target

    def transpile_modname_source_target(self, spec, modname, source, target):
        """
        The function that gets called by compile_transpile_entry for
        processing the provided JavaScript source file provided by some
        Python package through the transpiler instance.
        """

        if not isinstance(self.transpiler, BaseUnparser):
            _deprecation_warning(
                'transpiler callable assigned to %r must be an instance of '
                'calmjs.parse.unparsers.base.BaseUnparser by calmjs-4.0.0; '
                'if the original transpile behavior is to be retained, the '
                'subclass may instead override this method to call '
                '`simple_transpile_modname_source_target` directly, as '
                'this fallback behavior will be removed by calmjs-4.0.0' % (
                    self,
                )
            )
            return self.simple_transpile_modname_source_target(
                spec, modname, source, target)

        # do the new thing here.
        return self._transpile_modname_source_target(
            spec, modname, source, target)

    def _transpile_modname_source_target(self, spec, modname, source, target):
        bd_target = self._generate_transpile_target(spec, target)
        logger.info('Transpiling %s to %s', source, bd_target)
        reader = partial_open(source, 'r')
        writer_main = partial_open(bd_target, 'w')
        writer_map = (
            partial_open(bd_target + '.map', 'w')
            if spec.get(GENERATE_SOURCE_MAP) else
            None
        )
        write(self.transpiler, [
            read(self.parser, reader)], writer_main, writer_map)

    def simple_transpile_modname_source_target(
            self, spec, modname, source, target):
        """
        The original simple transpile method called by compile_transpile
        on each target.
        """

        opener = self.opener
        bd_target = self._generate_transpile_target(spec, target)
        logger.info('Transpiling %s to %s', source, bd_target)
        with opener(source, 'r') as reader, opener(bd_target, 'w') as _writer:
            writer = SourceWriter(_writer)
            self.transpiler(spec, reader, writer)
            if writer.mappings and spec.get(GENERATE_SOURCE_MAP):
                source_map_path = bd_target + '.map'
                with open(source_map_path, 'w') as sm_fd:
                    self.dump(encode_sourcemap(
                        filename=bd_target,
                        mappings=writer.mappings,
                        sources=[source],
                    ), sm_fd)

                # just use basename
                source_map_url = basename(source_map_path)
                _writer.write('\n//# sourceMappingURL=')
                _writer.write(source_map_url)
                _writer.write('\n')

    def compile_transpile_entry(self, spec, entry):
        """
        Handler for each entry for the transpile method of the compile
        process.  This invokes the transpiler that was set up to
        transpile the input files into the build directory.
        """

        modname, source, target, modpath = entry
        transpiled_modpath = {modname: modpath}
        transpiled_target = {modname: target}
        export_module_name = [modname]
        self.transpile_modname_source_target(spec, modname, source, target)
        return transpiled_modpath, transpiled_target, export_module_name

    def compile_bundle_entry(self, spec, entry):
        """
        Handler for each entry for the bundle method of the compile
        process.  This copies the source file or directory into the
        build directory.
        """

        modname, source, target, modpath = entry
        bundled_modpath = {modname: modpath}
        bundled_target = {modname: target}
        export_module_name = []
        if isfile(source):
            export_module_name.append(modname)
            copy_target = join(spec[BUILD_DIR], target)
            if not exists(dirname(copy_target)):
                makedirs(dirname(copy_target))
            shutil.copy(source, copy_target)
        elif isdir(source):
            copy_target = join(spec[BUILD_DIR], modname)
            shutil.copytree(source, copy_target)

        return bundled_modpath, bundled_target, export_module_name

    def compile_loaderplugin_entry(self, spec, entry):
        """
        Generic loader plugin entry handler.

        The default implementation assumes that everything up to the
        first '!' symbol resolves to some known loader plugin within
        the registry.

        The registry instance responsible for the resolution of the
        loader plugin handlers must be available in the spec under
        CALMJS_LOADERPLUGIN_REGISTRY
        """

        modname, source, target, modpath = entry
        handler = spec[CALMJS_LOADERPLUGIN_REGISTRY].get(modname)
        if handler:
            return handler(self, spec, modname, source, target, modpath)
        logger.warning(
            "no loaderplugin handler found for plugin entry '%s'", modname)
        return {}, {}, []

    # The naming methods, which are needed by certain toolchains that
    # need to generate specific names to maintain compatibility.  The
    # intended use case for this set of methods is to provide a rigidly
    # defined name handling ruleset for a given implementation of a
    # toolchain, but for toolchains that have its own custom naming
    # schemes per whatever value combination, further handling can be
    # done within each of the compile_* and/or compile_*_entry methods
    # that are enabled or registered for use for that particular
    # toolchain implementation.

    # Also note that 'source' and 'target' refer to 'sourcepath' and
    # 'targetpath' respectively in argument and method names.

    def modname_source_to_modname(self, spec, modname, source):
        """
        Method to get a modname.  Should really return the modname, but
        subclass has the option to override this.

        Called by generator method `_gen_modname_source_target_modpath`.
        """

        return modname

    def modname_source_to_source(self, spec, modname, source):
        """
        Method to get a source file name.  Should really return itself,
        but subclass has the option to override this.

        Called by generator method `_gen_modname_source_target_modpath`.
        """

        return source

    def modname_source_to_target(self, spec, modname, source):
        """
        Create a target file name from the input module name and its
        source file name.  The result should be a path relative to the
        build_dir, and this is derived directly from the modname with NO
        implicit convers of path separators (i.e. '/' or any other) into
        a system or OS specific form (e.g. '\\').  The rationale for
        this choice is that there exists Node.js/JavaScript tools that
        handle this internally and/or these paths and values are
        directly exposed on the web and thus these separators must be
        preserved.

        If the specific implementation requires this to be done,
        implementations may override by wrapping the result of this
        using os.path.normpath.  For the generation of transpile write
        targets, this will be done in _generate_transpile_target.

        Default is to append the module name with the filename_suffix
        assigned to this instance (setup by setup_filename_suffix), iff
        the provided source also end with this filename suffix.

        However, certain tools have issues dealing with loader plugin
        syntaxes showing up on the filesystem (and certain filesystems
        definitely do not like some of the characters), so the usage of
        the loaderplugin registry assigned to the spec may be used for
        lookup if available.

        Called by generator method `_gen_modname_source_target_modpath`.
        """

        loaderplugin_registry = spec.get(CALMJS_LOADERPLUGIN_REGISTRY)
        if '!' in modname and loaderplugin_registry:
            handler = loaderplugin_registry.get(modname)
            if handler:
                return handler.modname_source_to_target(
                    self, spec, modname, source)

        if (source.endswith(self.filename_suffix) and
                not modname.endswith(self.filename_suffix)):
            return modname + self.filename_suffix
        else:
            # assume that modname IS the filename
            return modname

    def modname_source_target_to_modpath(self, spec, modname, source, target):
        """
        Typical JavaScript tools will get confused if '.js' is added, so
        by default the same modname is returned as path rather than the
        target file for the module path to be written to the output file
        for linkage by tools.  Some other tools may desire the target to
        be returned instead, or construct some other string that is more
        suitable for the tool that will do the assemble and link step.

        The modname and source argument provided to aid pedantic tools,
        but really though this provides more consistency to method
        signatures.

        Called by generator method `_gen_modname_source_target_modpath`.
        """

        return modname

    def modname_source_target_modnamesource_to_modpath(
            self, spec, modname, source, target, modname_source):
        """
        Typical JavaScript tools will get confused if '.js' is added, so
        by default the same modname is returned as path rather than the
        target file for the module path to be written to the output file
        for linkage by tools.  Some other tools may desire the target to
        be returned instead, or construct some other string that is more
        suitable for the tool that will do the assemble and link step.

        The modname and source argument provided to aid pedantic tools,
        but really though this provides more consistency to method
        signatures.

        Same as `self.modname_source_target_to_modpath`, but includes
        the original raw key-value as a 2-tuple.

        Called by generator method `_gen_modname_source_target_modpath`.
        """

        return self.modname_source_target_to_modpath(
            spec, modname, source, target)

    # Generator methods

    def _gen_modname_source_target_modpath(self, spec, d):
        """
        Private generator that will consume those above functions.  This
        should NOT be overridden.

        Produces the following 4-tuple on iteration with the input dict;
        the definition is written at the module level documention for
        calmjs.toolchain, but in brief:

        modname
            The JavaScript module name.
        source
            Stands for sourcepath - path to some JavaScript source file.
        target
            Stands for targetpath - the target path relative to
            spec[BUILD_DIR] where the source file will be written to
            using the method that genearted this entry.
        modpath
            The module path that is compatible with tool referencing
            the target.  While this is typically identical with modname,
            some tools require certain modifications or markers in
            additional to what is presented (e.g. such as the addition
            of a '?' symbol to ensure absolute lookup).
        """

        for modname_source in d.items():
            try:
                modname = self.modname_source_to_modname(spec, *modname_source)
                source = self.modname_source_to_source(spec, *modname_source)
                target = self.modname_source_to_target(spec, *modname_source)
                modpath = self.modname_source_target_modnamesource_to_modpath(
                    spec, modname, source, target, modname_source)
            except ValueError as e:
                # figure out which of the above 3 functions failed by
                # acquiring the name from one frame down.
                f_name = sys.exc_info()[2].tb_next.tb_frame.f_code.co_name

                if isinstance(e, ValueSkip):
                    # a purposely benign failure.
                    log = partial(
                        logger.info,
                        "toolchain purposely skipping on '%s', "
                        "reason: %s, where modname='%s', source='%s'",
                    )
                else:
                    log = partial(
                        logger.warning,
                        "toolchain failed to acquire name with '%s', "
                        "reason: %s, where modname='%s', source='%s'; "
                        "skipping",
                    )

                log(f_name, e, *modname_source)
                continue
            yield modname, source, target, modpath

    # The core functions to be implemented for the toolchain.

    def prepare(self, spec):
        """
        Optional preparation step for handling the spec.

        Implementation can make use of this to do pre-compilation
        checking and/or other validation steps in order to result in a
        successful compilation run.
        """

    def compile(self, spec):
        """
        Generic step that compiles from a spec to build the specified
        things into the build directory `build_dir`, by gathering all
        the files and feed them through the transpilation process or by
        simple copying.
        """

        spec[EXPORT_MODULE_NAMES] = export_module_names = spec.get(
            EXPORT_MODULE_NAMES, [])
        if not isinstance(export_module_names, list):
            raise TypeError(
                "spec provided a '%s' but it is not of type list "
                "(got %r instead)" % (EXPORT_MODULE_NAMES, export_module_names)
            )

        def compile_entry(method, read_key, store_key):
            spec_read_key = read_key + self.sourcepath_suffix
            spec_modpath_key = store_key + self.modpath_suffix
            spec_target_key = store_key + self.targetpath_suffix

            if _check_key_exists(spec, [spec_modpath_key, spec_target_key]):
                logger.error(
                    "aborting compile step %r due to existing key", entry,
                )
                return

            sourcepath_dict = spec.get(spec_read_key, {})
            entries = self._gen_modname_source_target_modpath(
                spec, sourcepath_dict)
            (spec[spec_modpath_key], spec[spec_target_key],
                new_module_names) = method(spec, entries)
            logger.debug(
                "entry %r "
                "wrote %d entries to spec[%r], "
                "wrote %d entries to spec[%r], "
                "added %d export_module_names",
                entry,
                len(spec[spec_modpath_key]), spec_modpath_key,
                len(spec[spec_target_key]), spec_target_key,
                len(new_module_names),
            )
            export_module_names.extend(new_module_names)

        for entry in self.compile_entries:
            if isinstance(entry, ToolchainSpecCompileEntry):
                log = partial(
                    logging.getLogger(entry.logger).log,
                    entry.log_level,
                    (
                        entry.store_key + "%s['%s'] is being rewritten from "
                        "'%s' to '%s'; configuration may now be invalid"
                    ),
                ) if entry.logger else None
                compile_entry(partial(
                    toolchain_spec_compile_entries, self,
                    process_name=entry.process_name,
                    overwrite_log=log,
                ), entry.read_key, entry.store_key)
                continue

            m, read_key, store_key = entry
            if callable(m):
                method = m
            else:
                method = getattr(self, self.compile_prefix + m, None)
                if not callable(method):
                    logger.error(
                        "'%s' not a callable attribute for %r from "
                        "compile_entries entry %r; skipping", m, self, entry
                    )
                    continue

            compile_entry(method, read_key, store_key)

    def assemble(self, spec):
        """
        Assemble all the compiled files.

        This was intended to be the function that provides the
        aggregation of all compiled files in the build directory into a
        form that can then be linked.  Typically this is for the
        generation of an actual specification or instruction file that
        will be passed to the linker, which is some binary that is
        installed on the system.
        """

        raise NotImplementedError

    def link(self, spec):
        """
        Should pass in the manifest path to the final JS linker, which
        is typically the bundler.
        """

        raise NotImplementedError

    def finalize(self, spec):
        """
        Optional finalizing step, where further usage of the build_dir,
        scripts and/or results are needed.  This can be used to run some
        specific scripts through node's import system directly on the
        pre-linked assembled files, for instance.

        This step is optional.
        """

    def _calf(self, spec):
        """
        The main call, assuming the base spec is prepared.

        Also, no advices will be triggered.
        """

        self.prepare(spec)
        self.compile(spec)
        self.assemble(spec)
        self.link(spec)
        self.finalize(spec)

    def setup_apply_advice_packages(
            self, spec, default_advice_registry=CALMJS_TOOLCHAIN_ADVICE):
        """
        This method sets up the advices that have been specified in the
        ADVICE_PACKAGES key, and apply the advices to the spec.
        """

        advice_registry_key = spec.get(
            CALMJS_TOOLCHAIN_ADVICE_REGISTRY, default_advice_registry)
        advice_registry = get_registry(advice_registry_key)
        if not isinstance(advice_registry, AdviceRegistry):
            logger.warning(
                "registry key '%s' resulted in %r which is not a valid advice "
                "registry; all package advice steps will be skipped",
                advice_registry_key, advice_registry
            )
            return
        logger.debug(
            "setting up advices using %s '%s'",
            cls_to_name(type(advice_registry)), advice_registry_key
        )
        advice_registry.apply_toolchain_spec(self, spec)

    def calf(self, spec):
        """
        Typical safe usage is this, which sets everything that could be
        problematic up.

        Requires the filename which everything will be produced to.
        """

        if not isinstance(spec, Spec):
            raise TypeError('spec must be of type Spec')

        # The following ensure steps really should be formalised into
        # some form of setup.  This may be a version 4+ item to consider
        # for integration, where the current SETUP advice is changed to
        # BEFORE_SETUP and have the following ensure step be part of the
        # default setup.

        # ensure build directory is defined and sane.
        if not spec.get(BUILD_DIR):
            tempdir = realpath(mkdtemp())
            spec.advise(CLEANUP, shutil.rmtree, tempdir)
            build_dir = join(tempdir, 'build')
            mkdir(build_dir)
            spec[BUILD_DIR] = build_dir
        else:
            build_dir = self.realpath(spec, BUILD_DIR)
            if not isdir(build_dir):
                logger.error("build_dir '%s' is not a directory", build_dir)
                raise_os_error(errno.ENOTDIR, build_dir)

        # ensure export target is sane
        self.realpath(spec, EXPORT_TARGET)

        # ensure advices specific to packages are applied, and applied
        # using the advice to maintain the feature as it was when
        # initially implemented as part of the runtime.  This also allow
        # advice setup exceptions be handled as expected.
        spec.advise(SETUP, self.setup_apply_advice_packages, spec)

        try:
            # Finally, handle setup which may set up the deferred
            # advices, as all the toolchain (and its runtime and/or its
            # parent runtime and related toolchains) spec advises should
            # have been done.
            spec.handle(SETUP)

            process = ('prepare', 'compile', 'assemble', 'link', 'finalize')
            for p in process:
                spec.handle('before_' + p)
                getattr(self, p)(spec)
                spec.handle('after_' + p)
            spec.handle(SUCCESS)
        except ToolchainCancel:
            if spec.get(DEBUG):
                log_exc_reason(*sys.exc_info())
        except ToolchainAbort:
            if spec.get(DEBUG):
                log_exc_reason(*sys.exc_info())
            raise
        finally:
            spec.handle(CLEANUP)

    def __call__(self, spec):
        """
        Alias, also make this callable directly.
        """

        self.calf(spec)


class NullToolchain(Toolchain):
    """
    A null toolchain that does nothing except maybe move some files
    around.
    """

    def __init__(self):
        super(NullToolchain, self).__init__()

    def setup_transpiler(self):
        self.transpiler = null_transpiler

    def transpile_modname_source_target(self, spec, modname, source, target):
        """
        Calls the original version.
        """

        return self.simple_transpile_modname_source_target(
            spec, modname, source, target)

    def prepare(self, spec):
        """
        Does absolutely nothing
        """

        spec['prepare'] = 'prepared'

    def assemble(self, spec):
        """
        Does absolutely nothing
        """

        spec['assemble'] = 'assembled'

    def link(self, spec):
        """
        Does absolutely nothing
        """

        spec['link'] = 'linked'


class ES5Toolchain(Toolchain):
    """
    A null toolchain that does nothing except maybe move some files
    around, using the es5 Unparser, pretty printer version.
    """

    def __init__(self, *a, **kw):
        super(ES5Toolchain, self).__init__(*a, **kw)

    def setup_transpiler(self):
        self.transpiler = pretty_printer()
        self.parser = parse



# -*- coding: utf-8 -*-
import unittest
import os
import sys

from os.path import exists
from os.path import join
from os.path import normcase
from os.path import relpath
from os.path import sep

from types import ModuleType
import pkg_resources

from calmjs import indexer
from calmjs.utils import pretty_logging

from calmjs.testing.utils import make_dummy_dist
from calmjs.testing.utils import make_multipath_module3
from calmjs.testing.utils import stub_item_attr_value
from calmjs.testing.utils import mkdtemp
from calmjs.testing.mocks import StringIO

# default dummy entry point for calmjs.
calmjs_ep = pkg_resources.EntryPoint.parse('demo = demo')
calmjs_ep.dist = pkg_resources.working_set.find(
    pkg_resources.Requirement.parse('calmjs'))
calmjs_dist_dir = pkg_resources.resource_filename(
    calmjs_ep.dist.as_requirement(), '')


def to_os_sep_path(p):
    # turn the given / separated path into an os specific path
    return sep.join(p.split('/'))


def rp_calmjs(mapping):
    # helper to remap all path values in the provided mapping to
    # be relative of calmjs distribution location
    return {k: relpath(v, calmjs_dist_dir) for k, v in mapping.items()}


class PkgResourcesIndexTestCase(unittest.TestCase):
    """
    This series of tests muck with python module internals.
    """

    def setUp(self):
        # `dummyns.submod` emulates a package that also provide the
        # `dummyns` namespace package that got installed after the
        # other package `dummyns`.
        ds_egg_root = join(mkdtemp(self), 'dummyns.submod')
        dummyns_path = join(ds_egg_root, 'dummyns')
        dummyns = ModuleType('dummyns')
        dummyns.__file__ = join(dummyns_path, '__init__.py')
        dummyns.__path__ = [dummyns_path]
        self.addCleanup(sys.modules.pop, 'dummyns')
        sys.modules['dummyns'] = dummyns

        dummyns_submod_path = join(ds_egg_root, 'dummyns', 'submod')
        dummyns_submod = ModuleType('dummyns.submod')
        dummyns_submod.__file__ = join(dummyns_submod_path, '__init__.py')
        dummyns_submod.__path__ = [dummyns_submod_path]
        self.addCleanup(sys.modules.pop, 'dummyns.submod')
        sys.modules['dummyns.submod'] = dummyns_submod

        os.makedirs(dummyns_submod_path)

        with open(join(dummyns_path, '__init__.py'), 'w') as fd:
            fd.write('')

        with open(join(dummyns_submod_path, '__init__.py'), 'w') as fd:
            fd.write('')

        self.nested_res = join(dummyns_submod_path, 'data.txt')
        self.nested_data = 'data'
        with open(self.nested_res, 'w') as fd:
            fd.write(self.nested_data)

        # create the package proper
        self.dummyns_submod_dist = make_dummy_dist(self, ((
            'namespace_packages.txt',
            'dummyns\n'
            'dummyns.submod\n',
        ), (
            'entry_points.txt',
            '[dummyns.submod]\n'
            'dummyns.submod = dummyns.submod:attr\n',
        ),), 'dummyns.submod', '1.0', working_dir=ds_egg_root)

        self.ds_egg_root = ds_egg_root
        self.dummyns_path = dummyns_path

        self.mod_dummyns = dummyns
        self.mod_dummyns_submod = dummyns_submod

    def test_invalid_missing_entry_point(self):
        with self.assertRaises(AttributeError):
            indexer.resource_filename_mod_entry_point('dummyns', None)

    def test_missing_distribution(self):
        d_egg_root = join(mkdtemp(self), 'dummyns')
        make_dummy_dist(self, ((
            'namespace_packages.txt',
            'not_ns\n',
        ), (
            'entry_points.txt',
            '[dummyns]\n'
            'dummyns = dummyns:attr\n',
        ),), 'dummyns', '2.0', working_dir=d_egg_root)
        working_set = pkg_resources.WorkingSet([
            d_egg_root,
            self.ds_egg_root,
        ])
        dummyns_ep = next(working_set.iter_entry_points('dummyns'))
        with pretty_logging(stream=StringIO()) as fd:
            p = indexer.resource_filename_mod_entry_point(
                'dummyns', dummyns_ep)
        # not stubbed working_set, so this is derived using fallback
        # value from the sys.modules['dummyns'] location
        self.assertEqual(normcase(p), normcase(self.dummyns_path))
        self.assertIn("distribution 'dummyns 2.0' not found", fd.getvalue())

    def test_standard(self):
        d_egg_root = join(mkdtemp(self), 'dummyns')

        make_dummy_dist(self, ((
            'namespace_packages.txt',
            'dummyns\n',
        ), (
            'entry_points.txt',
            '[dummyns]\n'
            'dummyns = dummyns:attr\n',
        ),), 'dummyns', '1.0', working_dir=d_egg_root)
        working_set = pkg_resources.WorkingSet([
            d_egg_root,
            self.ds_egg_root,
        ])
        # ensure the working_set is providing the distributions being
        # mocked here so that resource_filename will resolve correctly
        stub_item_attr_value(self, pkg_resources, 'working_set', working_set)

        moddir = join(d_egg_root, 'dummyns')
        os.makedirs(moddir)

        # make this also a proper thing
        with open(join(moddir, '__init__.py'), 'w') as fd:
            fd.write('')

        dummyns_ep = next(working_set.iter_entry_points('dummyns'))
        p = indexer.resource_filename_mod_entry_point('dummyns', dummyns_ep)

        # finally, this should work.
        self.assertEqual(normcase(p), normcase(moddir))

    def test_relocated_distribution(self):
        root = mkdtemp(self)
        dummyns_path = join(root, 'dummyns')

        make_dummy_dist(self, ((
            'namespace_packages.txt',
            'dummyns\n',
        ), (
            'entry_points.txt',
            '[dummyns]\n'
            'dummyns = dummyns:attr\n',
        ),), 'dummyns', '1.0', working_dir=root)
        working_set = pkg_resources.WorkingSet([
            root,
            self.ds_egg_root,
        ])
        # activate this as the working set
        stub_item_attr_value(self, pkg_resources, 'working_set', working_set)
        dummyns_ep = next(working_set.iter_entry_points('dummyns'))
        with pretty_logging(stream=StringIO()) as fd:
            p = indexer.resource_filename_mod_entry_point(
                'dummyns', dummyns_ep)
        # since the actual location is not created)
        self.assertIsNone(p)
        self.assertIn("does not exist", fd.getvalue())

        # retry with the module directory created at the expected location
        os.mkdir(dummyns_path)
        with pretty_logging(stream=StringIO()) as fd:
            p = indexer.resource_filename_mod_entry_point(
                'dummyns', dummyns_ep)
        self.assertEqual(normcase(p), normcase(dummyns_path))
        self.assertEqual('', fd.getvalue())

    def test_nested_namespace(self):
        self.called = None

        def _exists(p):
            self.called = p
            return exists(p)

        working_set = pkg_resources.WorkingSet([
            self.ds_egg_root,
        ])
        stub_item_attr_value(self, pkg_resources, 'working_set', working_set)
        stub_item_attr_value(self, indexer, 'exists', _exists)

        dummyns_ep = next(working_set.iter_entry_points('dummyns.submod'))
        p = indexer.resource_filename_mod_entry_point(
            'dummyns.submod', dummyns_ep)
        self.assertEqual(p, self.called)

        with open(join(p, 'data.txt')) as fd:
            data = fd.read()

        self.assertEqual(data, self.nested_data)


class IndexerTestCase(unittest.TestCase):

    def test_register(self):
        def bar_something():
            "dummy method"

        registry = {'foo': {}, 'bar': {}}
        with self.assertRaises(TypeError):
            indexer.register('foo', registry=registry)(bar_something)

        indexer.register('bar', registry=registry)(bar_something)
        self.assertEqual(registry['bar']['something'], bar_something)

    def test_get_modpath_last_empty(self):
        module = ModuleType('nothing')
        with pretty_logging(stream=StringIO()) as fd:
            self.assertEqual(indexer.modpath_last(module, None), [])
        self.assertIn(
            "module 'nothing' does not appear to be a namespace module",
            fd.getvalue())

    def test_get_modpath_last_multi(self):
        module = ModuleType('nothing')
        module.__path__ = ['/path/to/here', '/path/to/there']
        self.assertEqual(
            indexer.modpath_last(module, None), ['/path/to/there'])

    def test_get_modpath_all_empty(self):
        module = ModuleType('nothing')
        with pretty_logging(stream=StringIO()) as fd:
            self.assertEqual(indexer.modpath_all(module, None), [])
        self.assertIn(
            "module 'nothing' does not appear to be a namespace module",
            fd.getvalue())

    def test_get_modpath_all_multi(self):
        module = ModuleType('nothing')
        module.__path__ = ['/path/to/here', '/path/to/there']
        self.assertEqual(
            indexer.modpath_all(module, None),
            ['/path/to/here', '/path/to/there'],
        )

    def test_get_modpath_pkg_resources_valid(self):
        from calmjs.testing import module3
        result = indexer.modpath_pkg_resources(module3, calmjs_ep)
        self.assertEqual(len(result), 1)
        self.assertTrue(result[0].endswith(
            to_os_sep_path('calmjs/testing/module3')))

    def test_get_modpath_pkg_resources_missing_path(self):
        with pretty_logging(stream=StringIO()) as fd:
            self.assertEqual([], indexer.modpath_pkg_resources(
                None, calmjs_ep))
            self.assertIn(
                "None does not appear to be a valid module", fd.getvalue())
        with pretty_logging(stream=StringIO()) as fd:
            module = ModuleType('nothing')
            self.assertEqual([], indexer.modpath_pkg_resources(
                module, calmjs_ep))

        err = fd.getvalue()
        self.assertIn(
            "module 'nothing' and entry_point 'demo = demo'", err)
        # the input is fetched using a working entry_point, after all
        self.assertIn("resource path resolved to be '" + calmjs_dist_dir, err)
        self.assertIn("but it does not exist", fd.getvalue())

    def test_get_modpath_pkg_resources_invalid(self):
        # fake both module and entry point, which will trigger an import
        # error exception internally that gets logged.
        module = ModuleType('nothing')
        ep = pkg_resources.EntryPoint.parse('nothing = nothing')
        with pretty_logging(stream=StringIO()) as fd:
            self.assertEqual([], indexer.modpath_pkg_resources(module, ep))
        self.assertIn("module 'nothing' could not be imported", fd.getvalue())

    def test_get_modpath_pkg_resources_missing(self):
        # fake just the entry point, but provide a valid module.
        nothing = ModuleType('nothing')
        nothing.__path__ = []
        self.addCleanup(sys.modules.pop, 'nothing')
        sys.modules['nothing'] = nothing
        ep = pkg_resources.EntryPoint.parse('nothing = nothing')
        with pretty_logging(stream=StringIO()) as fd:
            self.assertEqual([], indexer.modpath_pkg_resources(nothing, ep))
        self.assertIn(
            "fail to resolve the resource path for module 'nothing' and "
            "entry_point 'nothing = nothing'", fd.getvalue())

    def test_module1_loader_es6(self):
        from calmjs.testing import module1
        results = rp_calmjs(indexer.mapper_es6(module1, calmjs_ep))
        self.assertEqual(results, {
            'calmjs/testing/module1/hello':
                to_os_sep_path('calmjs/testing/module1/hello.js'),
        })

    def test_module1_loader_python(self):
        from calmjs.testing import module1
        results = rp_calmjs(indexer.mapper_python(module1, calmjs_ep))
        self.assertEqual(results, {
            'calmjs.testing.module1.hello':
                to_os_sep_path('calmjs/testing/module1/hello.js'),
        })

    def test_module2_recursive_es6(self):
        from calmjs.testing import module2
        results = rp_calmjs(indexer.mapper(
            module2, calmjs_ep, globber='recursive'))
        self.assertEqual(results, {
            'calmjs/testing/module2/index':
                to_os_sep_path('calmjs/testing/module2/index.js'),
            'calmjs/testing/module2/helper':
                to_os_sep_path('calmjs/testing/module2/helper.js'),
            'calmjs/testing/module2/mod/helper':
                to_os_sep_path('calmjs/testing/module2/mod/helper.js'),
        })

    def test_module3_multi_path_all(self):
        """
        For modules that have multiple paths.  This is typically caused
        by specifying a module that is typically used as a namespace for
        other Python modules.  Normally this can interfere with imports
        but as long as a module is produced and the multiple path
        modpath method is used, the 'all' mapper will fulfil the order.
        """

        # See setup method for how it's built.
        module, index_js = make_multipath_module3(self)

        def join_mod3(*a):
            # use the actual value provided by the dummy module (which
            # references the real version.
            from calmjs.testing import module3
            mod3_dir = module3.__path__[0]
            return join(mod3_dir, *a)

        results = indexer.mapper(
            module, calmjs_ep, modpath='all', globber='recursive')
        self.assertEqual(results, {
            'calmjs/testing/module3/index': index_js,
            'calmjs/testing/module3/math': join_mod3('math.js'),
            'calmjs/testing/module3/mod/index': join_mod3('mod', 'index.js'),
        })

    def test_module3_multi_path_pkg_resources(self):
        """
        With the usage of pkg_resources modpath, the extra paths must be
        available somehow to this framework, but given that this is not
        setup through the framework proper the custom path provided by
        the mocked module object will never be used.
        """

        module, index_js = make_multipath_module3(self)

        def join_mod3(*a):
            return join(calmjs_dist_dir, 'calmjs', 'testing', 'module3', *a)

        results = indexer.mapper(
            module, calmjs_ep, modpath='pkg_resources', globber='recursive')
        self.assertEqual(results, {
            'calmjs/testing/module3/math': join_mod3('math.js'),
            'calmjs/testing/module3/mod/index': join_mod3('mod', 'index.js'),
        })

    def test_module2_callables(self):
        from calmjs.testing import module2
        results = rp_calmjs(indexer.mapper(
            module2,
            calmjs_ep,
            globber=indexer.globber_recursive,
            modname=indexer.modname_python,
            modpath=indexer.modpath_pkg_resources,
        ))
        self.assertEqual(results, {
            'calmjs.testing.module2.index':
                to_os_sep_path('calmjs/testing/module2/index.js'),
            'calmjs.testing.module2.helper':
                to_os_sep_path('calmjs/testing/module2/helper.js'),
            'calmjs.testing.module2.mod.helper':
                to_os_sep_path('calmjs/testing/module2/mod/helper.js'),
        })
